from ...client.resources.guild import Guild
from ...const import MISSING


class _GuildEvents:
    """
    Acts as a lookup table for Gateway events relating to
    guilds from Discord.
    """

    @classmethod
    def lookup(cls, name: str, data: dict) -> dict | MISSING:
        match name:
            case "GUILD_CREATE":
                return Guild
            case "GUILD_UPDATE":
                ...
            case "GUILD_DELETE":
                ...
            case "GUILD_BAN_ADD":
                ...
            case "GUILD_BAN_REMOVE":
                ...
            case "GUILD_EMOJIS_UPDATE":
                ...
            case "GUILD_STICKERS_UPDATE":
                ...
            case "GUILD_INTEGRATIONS_UPDATE":
                ...
            case "GUILD_MEMBER_ADD":
                ...
            case "GUILD_MEMBER_REMOVE":
                ...
            case "GUILD_MEMBER_UPDATE":
                ...
            case "GUILD_MEMBERS_CHUNK":
                ...
            case "GUILD_ROLE_CREATE":
                ...
            case "GUILD_ROLE_UPDATE":
                ...
            case "GUILD_ROLE_DELETE":
                ...
            case "GUILD_SCHEDULED_EVENT_CREATE":
                ...
            case "GUILD_SCHEDULED_EVENT_UPDATE":
                ...
            case "GUILD_SCHEDULED_EVENT_DELETE":
                ...
            case "GUILD_SCHEDULED_EVENT_USER_ADD":
                ...
            case "GUILD_SCHEDULED_EVENT_USER_REMOVE":
                ...
            case _:
                ...



from .abc import Snowflake
from ...utils.converters import optional_c
from .guild import Member

from attrs import define, field
from datetime import datetime

__all__ = ("Voice", "VoiceState")


@define()
class VoiceState:
    """
    Represents the state of a user's VOIP from Discord.

    Attributes
    ----------
    guild_id : `Snowflake`, optional
        The guild ID this voice state is for.
    channel_id : `Snowflake`, optional
        The channel ID this user is connected to.
    user_id : `Snowflake`
        The user ID this voice state is for.
    member : `Member`, optional
        The guild member this voice state is for.
    session_id : `str`
        The session ID for this voice state.
    deaf : `bool`
        Whether this user is deafened by the server or not.
    mute : `bool`
        Whether this user is muted by the server or not.
    self_deaf : `bool`
        Whether this user is locally deafened or not.
    self_mute : `bool`
        Whether this user is muted by the server or not.
    self_stream : `bool`
        Whether this user is streaming using 'Go Live' or not.
    self_video : `bool`
        Whether this user's camera is enabled or not.
    suppress : `bool`
        Whether this user is muted by the current user or not.
    request_to_speak_timestamp : `datetime`, optional
        The time at which the user requested to speak, if present.
    """

    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The guild ID this voice state is for."""
    channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The channel ID this user is connected to."""
    user_id: str | Snowflake = field(converter=Snowflake)
    """The user ID this voice state is for."""
    member: dict | Member | None = field(converter=optional_c(Member), default=None)
    """The guild member this voice state is for."""
    session_id: str = field()
    """The session ID for this voice state."""
    deaf: bool = field()
    """Whether this user is deafened by the server."""
    mute: bool = field()
    """Whether this user is muted by the server."""
    self_deaf: bool = field()
    """Whether this user is locally deafened."""
    self_mute: bool = field()
    """Whether this user is locally muted."""
    self_stream: bool = field(default=False)
    """Whether this user is streaming using 'Go Live'."""
    self_video: bool = field()
    """Whether this user's camera is enabled."""
    suppress: bool = field()
    """Whether this user is muted by the current user."""
    request_to_speak_timestamp: datetime | str | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """The time at which the user requested to speak."""


Voice = VoiceState



"""
retux
An API wrapper built with good intentions.

---

This project is open-sourced! If you'd like to make a
contribution, feel free to make a pull request.

- GitHub repository: https://github.com/ffl0w/retux
- Documentation: https://retux.rtfd.io (or see `/examples`!)
"""
from .client import *  # noqa
from .api.http import *  # noqa
from .api.events import *  # noqa



from attrs import define, field

from ...const import MISSING

from .misc import TypingStart
from .message import _MessageEvents
from .guild import _GuildEvents


@define()
class _Event:
    """
    Represents the base information of a Gateway event from Discord.

    Attributes
    ----------
    _name : `str`
        The name of the Gateway event.
    _bot : `Bot`, optional
        The bot instance linked to the event. This is added on so
        that we can hook into the instance's calls for callbacks
        related to the event, as well as potential HTTP calls.

        When left blank, this event represents purely receive-only
        information with no possible sending traits.
    """

    _name: str = field()
    """The name of the Gateway event."""
    _bot: "Bot" = field(default=MISSING)  # noqa
    """
    The bot instance linked to the event. This is added on so
    that we can hook into the instance's calls for callbacks
    related to the event, as well as potential HTTP calls.

    When left blank, this event represents purely receive-only
    information with no possible sending traits.
    """


class _EventTable(_MessageEvents, _GuildEvents):
    """
    Stores events from the Gateway for potential use dispatching.
    """

    @classmethod
    def lookup(self, name: str, data: dict):
        if messages := _MessageEvents.lookup(name, data):
            return messages
        elif guilds := _GuildEvents.lookup(name, data):
            return guilds
        if name == "TYPING_START":
            return TypingStart



from .abc import *  # noqa
from .app_commands import *  # noqa
from .components import *  # noqa
from .user import *  # noqa
from .role import *  # noqa
from .channel import *  # noqa
from .webhook import *  # noqa
from .guild_scheduled_event import *  # noqa

# Resource modules to be added when all dependencies are added as well:
# from .application import * # noqa
# from .channel import * # noqa



from enum import IntEnum
from json import dumps, loads
from logging import getLogger
from sys import platform
from time import perf_counter
from typing import Any, Protocol

from attrs import asdict, define, field
from cattrs import structure_attrs_fromdict
from trio import open_nursery, sleep, Nursery
from trio_websocket import ConnectionClosed, WebSocketConnection, open_websocket_url

from .error import (
    InvalidToken,
    RateLimited,
    InvalidShard,
    RequiresSharding,
    InvalidIntents,
    DisallowedIntents,
)
from .events.abc import _Event, _EventTable
from .events.connection import HeartbeatAck, InvalidSession, Ready, Reconnect, Resumed

from ..client.flags import Intents
from ..client.resources.abc import Snowflake
from ..const import MISSING, NotNeeded, __gateway_url__

logger = getLogger(__name__)


@define()
class _GatewayMeta:
    """Represents metadata for a Gateway connection."""

    version: int = field()
    """The version of the Gateway."""
    encoding: str = field()
    """The encoding type on Gateway payloads."""
    compress: str | None = field(default=None)
    """The compression type on Gateway payloads."""
    heartbeat_interval: float | None = field(default=None)
    """The heartbeat used to keep a Gateway connection alive."""
    session_id: str | None = field(default=None)
    """The ID of an existent session, used for when resuming a lost connection."""
    seq: int | None = field(default=None)
    """The sequence number on an existent session."""


class _GatewayOpCode(IntEnum):
    """Represents a Gateway event's operation code."""

    DISPATCH = 0
    """An event has been received."""
    HEARTBEAT = 1
    """A handshake sent periodically to ensure an alive connection."""
    IDENTIFY = 2
    """An identification event to start a session with the initial handshake."""
    PRESENCE_UPDATE = 3
    """An event sent to update the client's presence."""
    VOICE_STATE_UPDATE = 4
    """An event sent to update the client's voice state."""
    RESUME = 6
    """An event sent to resume a previous connection prior to disconnect."""
    RECONNECT = 7
    """An event received informing the client to reconnect."""
    REQUEST_GUILD_MEMBERS = 8
    """An event sent to receive information on offline guild members."""
    INVALID_SESSION = 9
    """
    An event received informing the client the connection has been invalidated.
    `RECONNECT` ultimately follows this.
    """
    HELLO = 10
    """An event received on a successfully initiated connection."""
    HEARTBEAT_ACK = 11
    """An event received to acknowledge a `HEARTBEAT` sent."""


@define(slots=False)
class _GatewayPayload:
    """
    Represents a Gateway payload, signifying data for events.

    The `s` (`sequence`) and `t` (`name`) attributes will only have
    a value when:

    - `op` (`opcode`) is `DISPATCH`.
    - A `RESUME` call has been made. (specific to the former)
    """

    op: int | _GatewayOpCode = field(converter=int)
    """The opcode of the payload."""
    d: Any | None = field(default=None)
    """The payload's event data."""
    s: int | None = field(default=None)
    """The sequence number, used for resuming sessions and heartbeats."""
    t: str | None = field(default=None)
    """The name of the payload's event."""

    @property
    def opcode(self) -> int:
        """The opcode of the payload."""
        return self.op

    @property
    def data(self) -> Any | None:
        """The payload's event data"""
        return self.d

    @property
    def sequence(self) -> int | None:
        """The sequence number, used for resuming sessions and heartbeats."""
        return self.s

    @property
    def name(self) -> str | None:
        """The name of the payload's event."""
        return self.t


class GatewayProtocol(Protocol):
    def __init__(
        self,
        token: str,
        intents: Intents,
        *,
        version: int = 10,
        encoding: str = "json",
        compress: NotNeeded[str] = MISSING,
    ):
        ...

    async def connect(self):
        ...

    async def reconnect(self):
        ...

    async def request_guild_members(
        self,
        guild_id: Snowflake,
        *,
        query: NotNeeded[str] = MISSING,
        limit: NotNeeded[int] = MISSING,
        presences: NotNeeded[bool] = MISSING,
        user_ids: NotNeeded[Snowflake | list[Snowflake]] = MISSING,
        nonce: NotNeeded[str] = MISSING,
    ):
        ...

    async def update_voice_state(
        self,
        guild_id: Snowflake,
        channel_id: NotNeeded[Snowflake] = MISSING,
        self_mute: NotNeeded[bool] = MISSING,
        self_deaf: NotNeeded[bool] = MISSING,
    ):
        ...

    @property
    def latency(self) -> float:
        ...


class GatewayClient(GatewayProtocol):
    """
    Represents a connection to Discord's Gateway. Gateways are Discord's
    form of real-time communication over secure WebSockets. Clients will
    receive events and data over the Gateway they are connected to and
    send data over the REST API.

    Upon instantiation, the class will attempt to assign encoding
    and compression values. Please see the documentation when instantiating
    for more details.

    ---

    Attributes
    ----------
    token : `str`
        The bots token.
    intents : `Intents`
        The intents to connect with.
    _conn : `trio_websocket.WebSocketConnection`
        An instance of a connection to the Gateway.
    _meta : `_GatewayMeta`
        Metadata representing connection parameters for the Gateway.
    _tasks : `trio.Nursery`
        The tasks associated with the Gateway, for reconnection and heart-beating.
    _closed : `bool`
        Whether the Gateway connection is closed or not.
    _stopped : `bool`
        Whether the Gateway connection was forcefully stopped or not.
    _heartbeat_ack : `bool`
        Whether we've received the first heartbeat acknowledgement or not.
    _last_ack : `list[float]`
        The before/after time of the last Gateway event tracked. See `latency` for Gateway connection timing.
    _bots : `list[retux.Bot]`
        The bot instances used for dispatching events.
    """

    # TODO: Add sharding and presence changing.

    __slots__ = ("token", "intents", "_meta")
    token: str
    """The bots token."""
    intents: Intents
    """The intents to connect with."""
    _conn: WebSocketConnection = None
    """An instance of a connection to the Gateway."""
    _meta: _GatewayMeta
    """Metadata representing connection parameters for the Gateway."""
    _tasks: Nursery = None
    """The tasks associated with the Gateway, for reconnection and heart-beating."""
    _closed: bool = True
    """Whether the Gateway connection is closed or not."""
    _stopped: bool = False
    """Whether the Gateway connection was forcefully stopped or not."""
    _heartbeat_ack: bool = False
    """Whether we've received the first heartbeat acknowledgement or not."""
    _last_ack: list[float] = []
    """The before/after time of the last Gateway event tracked. See `latency` for Gateway connection timing."""
    _bots: list["Bot"] = []  # noqa
    """The bot instances used for dispatching events."""

    def __init__(
        self,
        token: str,
        intents: Intents,
        *,
        version: int = 10,
        encoding: str = "json",
        compress: str = None,
    ):
        """
        Creates a new connection to the Gateway.

        Parameters
        ----------
        token : `str`
            The bots token to connect with.
        intents : `Intents`
            The intents to connect with.
        version : `int`, optional
            The version of the Gateway to use. Defaults to version `10`.
        encoding : `str`, optional
            The type of encoding to use on payloads. Defaults to `json`.
        compress : `str`, optional
            The type of data compression to use on payloads. Defaults to none.
        """
        self.token = token
        self.intents = intents
        self._meta = _GatewayMeta(version=version, encoding=encoding, compress=compress)

    async def __aenter__(self):
        self._tasks = open_nursery()
        nursery = await self._tasks.__aenter__()
        nursery.start_soon(self.reconnect)
        nursery.start_soon(self._heartbeat)
        return self

    async def __aexit__(self, *exc):
        return await self._tasks.__aexit__(*exc)

    async def _receive(self) -> _GatewayPayload:
        """
        Receives the next incoming payload from the Gateway.

        Returns
        -------
        `_GatewayPayload`
            A class of the payload data.
        """

        # FIXME: our exception handling neglects other rejection
        # reasons. A more thorough analysis of trio_websocket
        # is necessary to have an extensible exception of our
        # own for clarifying connection loss.

        try:
            resp = await self._conn.get_message()
            json = loads(resp)
            return structure_attrs_fromdict(json, _GatewayPayload)
        except ConnectionClosed:
            logger.warning("The connection to Discord's Gateway has closed.")
            self._closed = True
            await self.reconnect()

    async def _send(self, payload: _GatewayPayload):
        """
        Sends a payload to the Gateway.

        Parameters
        ----------
        payload : `_GatewayPayload`
            The payload to send.
        """

        # TODO: implement the gateway rate limiting logic here.
        # the theory of this is to "queue" dispatched information
        # from the Gateway when we enter a rate limit.

        try:
            json = dumps(asdict(payload))
            resp = await self._conn.send_message(json)  # noqa
        except ConnectionClosed:
            logger.warning("The connection to Discord's Gateway has closed.")
            await self._conn.aclose()
            await self.reconnect()

    async def connect(self):
        """Connects to the Gateway and initiates a WebSocket state."""
        self._stopped = False
        self._last_ack = [perf_counter(), perf_counter()]

        # FIXME: this connection type will only work with JSON in mind.
        # if other compression or encoding types are supplied, they
        # will not be properly digested. This is only added so others
        # may modify their GatewayClient to their liking.

        async with open_websocket_url(
            f"{__gateway_url__}?v={self._meta.version}&encoding={self._meta.encoding}"
            f"{'' if self._meta.compress is None else f'&compress={self._meta.compress}'}"
        ) as self._conn:
            self._closed = bool(self._conn.closed)

            if self._stopped:
                await self._conn.aclose()
            if self._closed:
                try:
                    match self._conn.closed.code:
                        case 4004:
                            raise InvalidToken(
                                "Your bots token is invalid. (Make sure there's a value, or reset if needed.)"
                            )
                        case 4008:
                            raise RateLimited(
                                "Your bot is being Gateway rate limited. You will be reconnected."
                            )
                        case 4010:
                            raise InvalidShard(
                                "You provided an invalid shard. Make sure the shard is correct! (https://discord.dev/topics/gateway#sharding)"
                            )
                        case 4011:
                            raise RequiresSharding(
                                "Your bot requires sharding, please use autoshard=True."
                            )
                        case 4013:
                            raise InvalidIntents(
                                "You provided an invalid intent. Make sure your intent is a value! (Did you also miss a | for adding more than one?)"
                            )
                        case 4014:
                            raise DisallowedIntents(
                                "You provided an intent that your bot is not approved for. Make sure your bot is verified and/or has it enabled in the Developer Portal."
                            )
                        case _:
                            pass
                except RateLimited:
                    await self._conn.aclose()
                    await self.reconnect()

            while not self._closed:
                data = await self._receive()

                if data:
                    await self._track(data)

    async def reconnect(self):
        """Reconnects to the Gateway and re-initiates a WebSocket state."""
        self._closed = True
        self._heartbeat_ack = False

        if self._closed:
            await self.connect()
        else:
            logger.info("Told to reconnect, but did not need to.")

    async def _track(self, payload: _GatewayPayload):
        """
        Tracks data sent from the Gateway and interprets it.

        Parameters
        ----------
        payload : `_GatewayPayload`
            The payload being sent from the Gateway.
        """
        logger.debug(
            f"Tracking payload: {payload.opcode}/{_GatewayOpCode(payload.opcode).name}."
            f"{'' if payload.name is None else f' ({payload.name})'}"
        )

        match _GatewayOpCode(payload.opcode):
            case _GatewayOpCode.HELLO:
                if self._meta.session_id:
                    logger.debug("Prior connection found, trying to resume.")
                    await self._resume()
                else:
                    logger.debug("New connection found, identifying to the Gateway.")
                    await self._identify()
                    self._meta.heartbeat_interval = payload.data["heartbeat_interval"] / 1000
                    logger.debug(f"Heartbeat set to {self._meta.heartbeat_interval}ms.")
                    self._heartbeat_ack = True
                    logger.debug("Began the heartbeat process.")
            case _GatewayOpCode.HEARTBEAT_ACK:
                self._last_ack[1] = perf_counter()
                logger.debug(f"The heartbeat was acknowledged. (took {self.latency}ms.)")
                await self._dispatch("HEARTBEAT_ACK", HeartbeatAck, latency=self.latency)
                self._last_ack[0] = perf_counter()
            case _GatewayOpCode.INVALID_SESSION:
                logger.info(
                    "Our Gateway connection has suddenly invalidated. Checking reconnection status."
                )
                await self._dispatch("INVALID_SESSION", InvalidSession, payload.data)

                if bool(payload.data):
                    logger.debug(
                        "The Gateway has told us to reconnect. Resuming last known connection."
                    )
                    await self._resume()
                else:
                    logger.debug(
                        "The given connection cannot be reconnected to. Starting new connection."
                    )
                    self._meta.session_id = None
                    await self._conn.aclose()
                    await self.reconnect()
            case _GatewayOpCode.RECONNECT:
                logger.info("The Gateway has told us to reconnect. Resuming last known connection.")
                await self._dispatch("RECONNECT", Reconnect)
                await self._resume()
            case _GatewayOpCode.DISPATCH:
                if payload.name not in ["RESUMED", "READY"]:
                    resource = _EventTable.lookup(payload.name, payload.data)
                    await self._dispatch(payload.name, resource, **payload.data)
        match payload.name:
            case "RESUMED":
                logger.debug(
                    f"The connection was resumed. (session: {self._meta.session_id}, sequence: {self._meta.seq}"
                )
                await self._dispatch("RESUMED", Resumed, **payload.data)
            case "READY":
                self._meta.session_id = payload.data["session_id"]
                self._meta.seq = payload.sequence
                logger.debug(
                    f"The Gateway has declared a ready connection. (session: {self._meta.session_id}, sequence: {self._meta.seq}"
                )
                await self._dispatch("READY", Ready, **payload.data)

    async def _hook(self, bot: "Bot") -> object:  # noqa
        """
        Hooks the Gateway to a bot for event dispatching.

        ---

        The `bot` field allows for numerous bots in theory
        to be hooked onto with one `GatewayClient` process
        being ran, allowing for IPC pipes and more intuitive
        sharding.

        ---

        Parameters
        ----------
        bot : `retux.Bot`
            The bot instance to hook onto. This instance
            can be any bot instance for interchangeable
            handling of 1 main Gateway.
        """
        logger.debug("Hooking the bot into the Gateway.")
        self._bots.append(bot)

    async def _dispatch(
        self, _name: str, data: list[dict] | dict | _Event | MISSING, *args, **kwargs
    ):
        """
        Dispatches an event from the Gateway.

        ---

        "Dispatching" is when the Gateway sends the client
        information regarding an event non-relevant to
        the connection.

        The `*args` and `**kwargs` signature is for data sent
        in the form of an `_Event`. These are only to be used
        to fill in the actual data of the Gateway event, whereas
        `data` is purely the dataclass.

        ---

        Parameters
        ----------
        _name : `str`
            The name of the event.
        data : `dict`, `_Event`, `MISSING`
            The supplied payload data from the event.

            If a resource was not able to be found for
            the event called for, `MISSING` will be given.
        """
        logger.debug(f"Dispatching {_name}: {data if isinstance(data, dict) else kwargs}")

        for bot in self._bots:
            if isinstance(data, dict) or isinstance(data, MISSING):
                await bot._trigger(_name.lower(), data)
            else:
                await bot._trigger(
                    _name.lower(),
                    data(
                        _name.lower(),
                        bot if "id" in kwargs else MISSING,
                        *args,
                        **{k: v for k, v in kwargs.items() if not k.startswith("_")},
                    ),
                )

    async def _identify(self):
        """Sends an identification payload to the Gateway."""
        payload = _GatewayPayload(
            op=_GatewayOpCode.IDENTIFY.value,
            d={
                "token": self.token,
                "intents": self.intents.value,
                "properties": {"os": platform, "browser": "retux", "device": "retux"},
            },
        )
        logger.debug("Sending an identification payload to the Gateway.")
        await self._send(payload)

    async def _resume(self):
        """Sends a resuming payload to the Gateway."""
        payload = _GatewayPayload(
            op=_GatewayOpCode.RESUME.value,
            d={
                "token": self.token,
                "session_id": self._meta.session_id,
                "seq": self._meta.seq,
            },
        )
        logger.debug("Sending a resuming payload to the Gateway.")
        await self._send(payload)

    async def _heartbeat(self):
        """Sends a heartbeat payload to the Gateway."""
        payload = _GatewayPayload(op=_GatewayOpCode.HEARTBEAT.value, d=self._meta.seq)

        # FIXME: Move towards a better solution for the heartbeat acknowledgement loop.
        # This is a really bad approach to fixing heartbeat timing, but this only fires
        # during the initialisation of the async context manager.
        # Please spare me Bluenix.

        logger.debug("Waiting the appropriate time for probable connection.")
        await sleep(1)

        while self._heartbeat_ack:
            logger.debug("Sending a heartbeat payload to the Gateway.")
            await self._send(payload)
            await sleep(self._meta.heartbeat_interval)

    async def request_guild_members(
        self,
        guild_id: Snowflake,
        *,
        query: NotNeeded[str] = MISSING,
        limit: NotNeeded[int] = MISSING,
        presences: NotNeeded[bool] = MISSING,
        user_ids: NotNeeded[Snowflake | list[Snowflake]] = MISSING,
        nonce: NotNeeded[str] = MISSING,
    ):
        """
        Sends a request for all guild members to the Gateway.

        Parameters
        ----------
        guild_id : `Snowflake`
            The ID of the guild to request from.
        query : `str`, optional
            The name of the guild member(s). If you're looking to
            receive all members of a guild, this is left untouched.
        limit : `int`, optional
            How many guild members you wish to return. When `query`
            is specified, only a maximum of `100` are returned.
            This should be left untouched with `query` for all
            members of a guild.
        presences : `bool`, optional
            Whether you only want to receive guild members with
            a presence. The `GUILD_PRESENCES` intent must be
            enabled in order to use.
        user_ids : `Snowflake` or `list[Snowflake]`, optional
            The IDs of members in the guild to return. This
            may be used in conjunction to `query`, and poses the
            same maximum as `limit` regardless of declaration.
        nonce : `str`, optional
            A nonce used for identification when receiving a
            `Guild Members Chunk` event.
        """
        payload = _GatewayPayload(
            op=_GatewayOpCode.REQUEST_GUILD_MEMBERS,
            d={
                "guild_id": guild_id,
                "query": "" if query is MISSING else query,
                "limit": 0 if limit is MISSING else limit,
            },
        )

        if presences is not MISSING:
            payload.data["presences"] = presences
        if user_ids is not MISSING:
            payload.data["user_ids"] = user_ids
        if nonce is not MISSING:
            payload.data["nonce"] = nonce

        logger.debug("Sending a payload requesting for guild members to the Gateway.")
        await self._send(payload)

    async def update_voice_state(
        self,
        guild_id: Snowflake,
        channel_id: NotNeeded[Snowflake] = MISSING,
        self_mute: NotNeeded[bool] = MISSING,
        self_deaf: NotNeeded[bool] = MISSING,
    ):
        """
        Sends a request updating the bots voice state to the Gateway.

        Parameters
        ----------
        guild_id : `Snowflake`
            The ID of the guild to request from.
        channel_id : `Snowflake`, optional
            The channel ID of the guild to update in.
            If the bot is trying to disconnect, this should
            be left untouched.
        self_mute : `bool`, optional
            Whether the bot is muting itself or not.
            Defaults to `False`.
        self_deaf : `bool`, optional
            Whether the bot is deafening itself or not.
            Defaults to `False`.
        """
        payload = _GatewayPayload(
            op=_GatewayOpCode.VOICE_STATE_UPDATE,
            d={
                "guild_id": guild_id,
                "channel_id": None if channel_id is MISSING else channel_id,
                "self_mute": False if self_mute is MISSING else self_mute,
                "self_deaf": False if self_deaf is MISSING else self_deaf,
            },
        )
        logger.debug("Sending a payload requesting a voice state update to the Gateway.")
        await self._send(payload)

    @property
    def latency(self) -> float:
        """
        The calculated difference between the last known set
        of acknowledgements for a Gateway event.
        """
        return self._last_ack[1] - self._last_ack[0]



from datetime import datetime
from enum import IntEnum, IntFlag
from attrs import define, field
from retux.client.resources.abc import Snowflake, Object, Partial
from .application import Application
from .sticker import Sticker, StickerItem
from .user import User
from .role import Role
from ...utils.converters import optional_c, list_c
from ..mixins import Respondable

__all__ = (
    "Channel",
    "FollowedChannel",
    "MessageActivity",
    "ChannelType",
    "ChannelFlags",
    "MessageType",
    "MessageActivityType",
    "MessageFlags",
    "VideoQualityMode",
    "Reaction",
    "Overwrite",
    "MessageActivity",
    "MessageReference",
    "ChannelMention",
    "ThreadMetadata",
    "ThreadMember",
    "_EmbedMedia",
    "_EmbedProvider",
    "_EmbedAuthor",
    "_EmbedFooter",
    "_EmbedField",
    "Embed",
    "Attachment",
    "Message",
)


class ChannelType(IntEnum):
    """
    Represents the types of channels from Discord

    Constants
    ---------
    GUILD_TEXT
        A text channel within a server.
    DM
        A direct message between users.
    GUILD_VOICE
        A voice channel within a server.
    GROUP_DM
        A direct message between multiple users.
    GUILD_CATEGORY
        An organizational category that contains up to 50 channels.
    GUILD_NEWS
        A channel that users can follow and crosspost into their own server.
    GUILD_NEWS_THREAD
        A temporary sub-channel within a GUILD_NEWS channel.
    GUILD_PUBLIC_THREAD
        A temporary sub-channel within a GUILD_TEXT channel.
    GUILD_PRIVATE_THREAD
        A temporary sub-channel within a GUILD_TEXT channel.

        Only viewable by those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE
        A voice channel for hosting events with an audience.
    GUILD_DIRECTORY
        The channel in a hub containing the listed servers.
    GUILD_FORUM
        A channel that can only contain threads.
    """

    GUILD_TEXT = 0
    """A text channel within a server."""
    DM = 1
    """A direct message between users."""
    GUILD_VOICE = 2
    """A voice channel within a server."""
    GROUP_DM = 3
    """A direct message between multiple users."""
    GUILD_CATEGORY = 4
    """An organizational category that contains up to 50 channels."""
    GUILD_NEWS = 5
    """A channel that users can follow and crosspost into their own server."""
    GUILD_NEWS_THREAD = 10
    """A temporary sub-channel within a GUILD_NEWS channel."""
    GUILD_PUBLIC_THREAD = 11
    """A temporary sub-channel within a GUILD_TEXT channel."""
    GUILD_PRIVATE_THREAD = 12
    """
    A temporary sub-channel within a GUILD_TEXT channel.

    Only viewable by those invited and those with the MANAGE_THREADS permission
    """
    GUILD_STAGE_VOICE = 13
    """A voice channel for hosting events with an audience."""
    GUILD_DIRECTORY = 14
    """The channel in a hub containing the listed servers."""
    GUILD_FORUM = 15
    """A channel that can only contain threads."""


class VideoQualityMode(IntEnum):
    """
    Represents video quality modes from Discord

    Constants
    ---------
    AUTO
        Discord chooses the quality for optimal performance.
    FULL
        720p video resolution (1280x720).
    """

    AUTO = 1
    """Discord chooses the quality for optimal performance."""
    FULL = 2
    """720p video resolution (1280x720)."""


class ChannelFlags(IntFlag):
    """The bitwise values that represent channel flags from Discord"""

    PINNED = 1 << 1
    """
    This thread is pinned to the top of its parent channel.

    Only available on forum channels.
    """


class MessageType(IntEnum):
    """
    Represnts message types from Discord

    Constants
    ---------
    DEFAULT
        A normal message.
    RECIPIENT_ADD
        Unknown
    RECIPIENT_REMOVE
        Unknown
    CALL
        An indicator for a new voice call in a DM.
    CHANNEL_NAME_CHANGE
        A notification message for name of a channel changing.
    CHANNEL_ICON_CHANGE
        A notification message for the icon of a DM changing.
    CHANNEL_PINNED_MESSAGE
        A notification message for a new pinned message.
    USER_JOIN
        A notification message about the name of a channel changing.
    GUILD_BOOST
        A notification message about a new server boost(s).
    GUILD_BOOST_TIER_1
        A notification message about a guild reaching level 1 boost perks.
    GUILD_BOOST_TIER_2
        A notification message about a guild reaching level 2 boost perks.
    GUILD_BOOST_TIER_3
        A notification message about a guild reaching level 3 boost perks.
    CHANNEL_FOLLOW_ADD
        A notification message about a new follower of a news channel.
    GUILD_DISCOVERY_DISQUALIFIED
        A notification message about a guild being disqualified for discovery.
    GUILD_DISCOVERY_REQUALIFIED
        A notification message about a guild being requalified for discovery.
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING
        An intial warning for a discovery grace period.
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING
        A final warning for a discovery grace period.
    THREAD_CREATED
        A notification message for a new thread in a channel.
    REPLY
        A reply to a message.
    CHAT_INPUT_COMMAND
        A resulting message from a slash command being used.
    THREAD_STARTER_MESSAGE
        A starter message for a thread.
    GUILD_INVITE_REMINDER
        A reminder about a guild invite.
    CONTEXT_MENU_COMMAND
        A resulting message from a context menu command being used.
    AUTO_MODERATION_ACTION
        A message from AutoMod describing an action it took.
    """

    DEFAULT = 0
    """A normal message."""
    RECIPIENT_ADD = 1
    RECIPIENT_REMOVE = 2
    CALL = 3
    """An indicator for a new voice call in a DM."""
    CHANNEL_NAME_CHANGE = 4
    """A notification message for name of a channel changing."""
    CHANNEL_ICON_CHANGE = 5
    """A notification message for the icon of a DM changing."""
    CHANNEL_PINNED_MESSAGE = 6
    """A notification message for a new pinned message."""
    USER_JOIN = 7
    """A notification message about the name of a channel changing."""
    GUILD_BOOST = 8
    """A notification message about a new server boost(s)."""
    GUILD_BOOST_TIER_1 = 9
    """A notification message about a guild reaching level 1 boost perks."""
    GUILD_BOOST_TIER_2 = 10
    """A notification message about a guild reaching level 2 boost perks."""
    GUILD_BOOST_TIER_3 = 11
    """A notification message about a guild reaching level 3 boost perks."""
    CHANNEL_FOLLOW_ADD = 12
    """A notification message about a new follower of a news channel."""
    GUILD_DISCOVERY_DISQUALIFIED = 14
    """A notification message about a guild being disqualified for discovery."""
    GUILD_DISCOVERY_REQUALIFIED = 15
    """A notification message about a guild being requalified for discovery."""
    GUILD_DISCOVERY_GRACE_PERIOD_INITIAL_WARNING = 16
    """An intial warning for a discovery grace period."""
    GUILD_DISCOVERY_GRACE_PERIOD_FINAL_WARNING = 17
    """A final warning for a discovery grace period."""
    THREAD_CREATED = 18
    """A notification message for a new thread in a channel."""
    REPLY = 19
    """A reply to a message."""
    CHAT_INPUT_COMMAND = 20
    """A resulting message from a slash command being used."""
    THREAD_STARTER_MESSAGE = 21
    """A starter message for a thread."""
    GUILD_INVITE_REMINDER = 22
    """A reminder about a guild invite."""
    CONTEXT_MENU_COMMAND = 23
    """A resulting message from a context menu command being used."""
    AUTO_MODERATION_ACTION = 24
    """A message from AutoMod describing an action it took."""


class MessageActivityType(IntEnum):
    """
    Represents a message activity type from Discord.

    Constants
    ---------
    JOIN
        A join message activity.
    SPECTATE
        A spectate message activity.
    LISTEN
        A listen message activity.
    JOIN_REQUEST
        A join request message activity.
    """

    JOIN = 1
    """A join message activity."""
    SPECTATE = 2
    """A spectate message activity."""
    LISTEN = 3
    """A listen message activity."""
    JOIN_REQUEST = 5
    """A join request message activity."""


class MessageFlags(IntFlag):
    """The bitwise values that represent message flags from Discord"""

    CROSSPOSTED = 1 << 0
    """This message has been published to subscribed channels."""
    IS_CROSSPOST = 1 << 1
    """This message originated from a message in another channel."""
    SUPPRESS_EMBEDS = 1 << 2
    """Do not include any embeds when serializing this message."""
    SOURCE_MESSAGE_DELETED = 1 << 3
    """The source message for this crosspost has been deleted."""
    URGENT = 1 << 4
    """This message came from the urgent message system."""
    HAS_THREAD = 1 << 5
    """This message has an associated thread, with the same ID as the message."""
    EPHEMERAL = 1 << 6
    """This message is only visible to the user who invoked the interaction."""
    LOADING = 1 << 7
    """This message is an interaction response showing that the bot is "thinking"."""
    FAILED_TO_MENTION_SOME_ROLES_IN_THREAD = 1 << 8
    """This message failed to mention some roles and add their members to the thread."""


@define(kw_only=True)
class Reaction:
    """
    Represents a reaction to a message from Discord.

    Attributes
    ----------
    count : `int`
        Number of reactions using the emoji.
    me : `bool`
        Whether or not the current user has reacted using this emoji.
    emoji : `EmojiPartial`
        Information about the emoji of the reaction.
    """

    count: int = field()
    """Number of reactions using the emoji."""
    me: bool = field()
    """Whether or not the current user has reacted using this emoji."""
    # TODO: Implement Partial Emoji object.
    # emoji: PartialEmoji = field(converter=PartialEmoji)
    """Information about the emoji of the reaction."""


@define(kw_only=True)
class Overwrite(Object):
    """
    Represents a permission overwrite from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the role or user.
    type : `int`
        The type of overwrite.

        Use `0` for to overwrite a role's permission and `1` to overwrite a guild member's permission.
    allow : `str`
        The bit set representing the overwrite's allowed permissions.
    deny : `str`
        The bit set representing the overwrite's denied permissions.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the role or user."""
    type: int = field()
    """
    The type of overwrite.

    Use `0` for to overwrite a role's permission and `1` to overwrite a guild member's permission.
    """
    allow: str = field()
    """The bit set representing the overwrite's allowed permissions."""
    deny: str = field()
    """The bit set representing the overwrite's denied permissions."""


@define(kw_only=True)
class MessageActivity:
    """
    Represents a message activity structure from Discord.

    Attributes
    ----------
    type : `int`
        The type of message activity.
    party_id : `Snowflake`, optional
        The ID of a party from a rich presence event.
    """

    type: int | MessageActivityType = field(converter=MessageActivityType)
    """The type of message activity."""
    party_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of a party from a rich presence event."""


@define(kw_only=True)
class MessageReference:
    """
    Represents a reference to a message from Discord.

    Attributes
    ----------
    message_id : `Snowflake`, optional
        The ID of the originating message.
    channel_id : `Snowflake`, optional
        The ID of the originating message's channel.
    guild_id : `Snowflake`, optional
        The ID of the originating message's guild.
    fail_if_not_exists : `bool`
        Whether or not to error if the referenced doesn't exist.

        If `False`, sends as a normal, non-reply message.
    """

    message_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the originating message."""
    channel_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the originating message's channel."""
    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the originating message's guild."""
    fail_if_not_exists: bool = field(default=True)
    """
    Whether or not to error if the referenced doesn't exist.

    If `False`, sends as a normal, non-reply message.
    """


@define(kw_only=True)
class FollowedChannel:
    """
    Represents a followed channel from Discord.

    Attributes
    ----------
    channel_id : `Snowflake`
        The ID of the source channel.
    webhook_id : `Snowflake`
        The ID of the created target webhook.
    """

    channel_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the source channel."""
    webhook_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the created target webhook."""


@define(kw_only=True)
class ChannelMention(Object):
    """
    Represents a channel mentioned in a message from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel being mentioned.
    guild_id : `Snowflake`
        The ID of the guild containing the channel being mentioned.
    type : `ChannelType`
        The type of the channel being mentioned.
    name : `str`
        The name of the channel being mentioned.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel being mentioned."""
    guild_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the guild containing the channel being mentioned."""
    type: int | ChannelType = field(converter=ChannelType)
    """The type of the channel being mentioned."""
    name: str = field()
    """The name of the channel being mentioned."""


@define(kw_only=True)
class ThreadMetadata:
    """
    Represents a thread metadata structure from Discord.

    Attributes
    ----------
    archived : `bool`
        Whether or not the thread is currently archived.
    auto_archive_duration : `int`
        The amount of minutes after the last message was sent before
        Discord automatically archives the thread.
    archive_timestamp : `datetime`
        When the thread's archive status last changed, used for
        calculating recent activity.
    locked : `bool`
        Whether or not the thread is locked.

        If a thread is locked, only users with MANAGE_THREADS permissions
        can unarchive it.
    invitable : `bool`
        Whether or not non-moderators can add other non-moderators to a thread.

        Only availible an private threads.
    create_timestamp: `datetime`, optional
        When the thread was created.

        Always populated unless the thread was created.
    """

    archived: bool = field()
    """Whether or not the thread is currently archived."""
    auto_archive_duration: int = field()
    """
    The amount of minutes after the last message was sent before
    Discord automatically archives the thread.
    """
    archive_timestamp: str | datetime = field(converter=datetime.fromisoformat)
    """When the thread's archive status last changed, used for calculating recent activity."""
    locked: bool = field()
    """
    Whether or not the thread is locked.

    If a thread is locked, only users with MANAGE_THREADS permissions can unarchive it.
    """
    invitable: bool = field()
    """
    Whether or not non-moderators can add other non-moderators to a thread.

    Only availible an private threads.
    """
    create_timestamp: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """
    When the thread was created.

    Always populated unless the thread was created.
    """


@define(kw_only=True)
class ThreadMember(Object):
    """
    Represents a thread member structure from Discord.

    Attributes
    ----------
    id : `Snowflake`, optional
        The ID of the thread.

        This field is omitted on the member sent within each thread in GUILD_CREATE.
    user_id : `Snowflake`
        The ID of the user.

        This field is omitted on the member sent within each thread in GUILD_CREATE.
    join_timestamp : `datetime`
        When the user joined the thread.
    flags : `int`
        The flags for any user-thread settings.

        This is not needed by the bot; it is only used for client notifications.
    """

    id: str | Snowflake | None = field(converter=optional_c(Snowflake))
    """
    The ID of the thread.

    This field is omitted on the member sent within each thread in GUILD_CREATE.
    """
    user_id: str | Snowflake = field(converter=Snowflake)
    """
    The ID of the user.

    This field is omitted on the member sent within each thread in GUILD_CREATE.
    """
    join_timestamp: str | datetime = field(converter=datetime.fromisoformat)
    """When the user joined the thread."""
    flags: int = field()
    """
    The flags for any user-thread settings.

    This is not needed by the bot; it is only used for client notifications.
    """


@define(kw_only=True)
class TextChannel(Partial, Object):
    """
    Represents a text channel from Discord.

    ---

    This is an event-specific dataclass that is not passed by
    Discord's Gateway. This will only contain the data relevant
    to a text channel -- and just the `id` of it. `type` is
    dropped as this is already given for the event handler via.
    specification in the typehinting.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    topic : `str`, optional
        The topic of the channel.

        A channel topic is in-between 1-1024 characters.
    nsfw : `bool`, optional
        Whether or not the channel is NSFW. Defaults to `False`.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    application_id : `Snowflake`, optional
        The ID of the application that created the dm if it is bot-created.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    last_pin_timestamp : `str`, optional
        The time when the last message was pinned.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel."""
    guild_id: str | Snowflake | None = field(default=None, converter=Snowflake)
    """
    The ID of the guild.

    This is nullable due to some Gateway events
    lacking the data for the ID.
    """
    position: int | None = field(default=None)
    """Sorted position of the channel."""
    permission_overwrites: list[dict] | list[Overwrite] | None = field(default=None)
    """Explicit permission overwrites for members and roles."""
    name: str | None = field(default=None)
    """
    The name of the channel.

    A channel name is in-between 1-100 characters.
    """
    topic: str | None = field(default=None)
    """
    The topic of the channel.

    A channel topic is in-between 1-1024 characters.
    """
    nsfw: bool | None = field(default=False)
    """Whether or not the channel is NSFW. Defaults to `False`."""
    last_message_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the last message sent in this channel.

    Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    """
    rate_limit_per_user: int | None = field(default=None)
    """
    Amount of seconds a user has to wait before sending another message

    Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    """
    recipients: list[dict] | list[User] | None = field(default=None)
    """The recipients of the dm."""
    icon: str | None = field(default=None)
    """The hash for the channel's icon."""
    owner_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the creator of the group dm or thread."""
    application_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the application that created the dm if it is bot-created."""
    parent_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the parent of the channel

    Represents the parent category for regular channels and the parent channel for threads.
    """
    last_pin_timestamp: str | None = field(default=None)
    """The time when the last message was pinned."""
    """The video quality mode of the voice channel."""
    message_count: int | None = field(default=None)
    """"
    The approximated amount of messages in a thread.

    Stops counting at `50`.
    """
    member_count: int | None = field(default=None)
    """
    The approximated amount of users in a thread.

    Stops counting at `50`.
    """
    # TODO: implement Thread Metadata object.
    thread_metadata: dict | ThreadMetadata | None = field(  # noqa
        converter=ThreadMetadata, default=None  # noqa
    )  # noqa
    """Thread-specific fields not needed by other channels."""
    # TODO: implement Thread Member object.
    member: dict | ThreadMember | None = field(converter=ThreadMember, default=None)  # noqa
    """
    The thread member representation of the user if they have joined the thread.

    This is only included on certain api endpoints.
    """
    default_auto_archive_duration: int | None = field(default=None)
    """
    The default archive duration for threads in minutes.

    Can be set to `60`, `1440`, `4320`, `10080`.
    """
    permissions: str | None = field(default=None)
    """
    The computed permissions for the invoking user in the channel, including any overwrites.

    Only included when part of the resolved data received on a slash command interaction.
    """
    flags: int | ChannelFlags | None = field(converter=ChannelFlags, default=None)
    """Channel flags combined as a bitfield."""


@define(kw_only=True)
class AnnouncementChannel(TextChannel):
    """
    Represents an announcement channel on Discord.

    ---

    This information is essentially in vogue the same as a
    text channel, presented with the `TextChannel` object.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    topic : `str`, optional
        The topic of the channel.

        A channel topic is in-between 1-1024 characters.
    nsfw : `bool`, optional
        Whether or not the channel is NSFW. Defaults to `False`.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    application_id : `Snowflake`, optional
        The ID of the application that created the dm if it is bot-created.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    last_pin_timestamp : `str`, optional
        The time when the last message was pinned.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """


@define(kw_only=True)
class ForumChannel(TextChannel):
    """
    Represents a forum channel on Discord.

    ---

    This information is essentially in vogue the same as a
    text channel, presented with the `TextChannel` object.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    topic : `str`, optional
        The topic of the channel.

        A channel topic is in-between 1-1024 characters.
    nsfw : `bool`, optional
        Whether or not the channel is NSFW. Defaults to `False`.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    application_id : `Snowflake`, optional
        The ID of the application that created the dm if it is bot-created.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    last_pin_timestamp : `str`, optional
        The time when the last message was pinned.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """


@define(kw_only=True)
class DMChannel(TextChannel):
    """
    Represents a DM channel on Discord.

    ---

    This information is essentially in vogue the same as a
    text channel, presented with the `TextChannel` object.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    topic : `str`, optional
        The topic of the channel.

        A channel topic is in-between 1-1024 characters.
    nsfw : `bool`, optional
        Whether or not the channel is NSFW. Defaults to `False`.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    application_id : `Snowflake`, optional
        The ID of the application that created the dm if it is bot-created.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    last_pin_timestamp : `str`, optional
        The time when the last message was pinned.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """


@define(kw_only=True)
class VoiceChannel(Partial, Object):
    """
    Represents a voice channel from Discord.

    ---

    This is an event-specific dataclass that is not passed by
    Discord's Gateway. This will only contain the data relevant
    to a voice channel -- and just the `id` of it. `type` is
    dropped as this is already given for the event handler via.
    specification in the typehinting.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    nsfw : `bool`, optional
        Whether or not the channel is NSFW. Defaults to `False`.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    bitrate : `int`, optional
        The bitrate of the voice channel.
    user_limit : `int`, optional
        The user limit of the voice channel.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    application_id : `Snowflake`, optional
        The ID of the application that created the dm if it is bot-created.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    last_pin_timestamp : `str`, optional
        The time when the last message was pinned.
    rtc_region : `str`, optional
        The channel's voice region ID if present, set to automatic when left as `None`..
    video_quality_mode : `VideoQualityMode`, optional
        The video quality mode of the voice channel.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel."""
    guild_id: str | Snowflake | None = field(default=None, converter=Snowflake)
    """
    The ID of the guild.

    This is nullable due to some Gateway events
    lacking the data for the ID.
    """
    position: int | None = field(default=None)
    """Sorted position of the channel."""
    permission_overwrites: list[dict] | list[Overwrite] | None = field(default=None)
    """Explicit permission overwrites for members and roles."""
    name: str | None = field(default=None)
    """
    The name of the channel.

    A channel name is in-between 1-100 characters.
    """
    nsfw: bool | None = field(default=False)
    """Whether or not the channel is NSFW. Defaults to `False`."""
    last_message_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the last message sent in this channel.

    Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    """
    bitrate: int | None = field(default=None)
    """The bitrate of the voice channel."""
    user_limit: int | None = field(default=None)
    """The user limit of the voice channel."""
    rate_limit_per_user: int | None = field(default=None)
    """
    Amount of seconds a user has to wait before sending another message

    Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    """
    icon: str | None = field(default=None)
    """The hash for the channel's icon."""
    owner_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the creator of the group dm or thread."""
    application_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the application that created the dm if it is bot-created."""
    parent_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the parent of the channel

    Represents the parent category for regular channels and the parent channel for threads.
    """
    last_pin_timestamp: str | None = field(default=None)
    """The time when the last message was pinned."""
    rtc_region: str | None = field(default=None)
    """The channel's voice region ID if present, set to automatic when left as `None`."""
    video_quality_mode: int | VideoQualityMode | None = field(
        converter=VideoQualityMode, default=None
    )
    """The video quality mode of the voice channel."""
    message_count: int | None = field(default=None)
    """"
    The approximated amount of messages in a thread.

    Stops counting at `50`.
    """
    member_count: int | None = field(default=None)
    """
    The approximated amount of users in a thread.

    Stops counting at `50`.
    """
    # TODO: implement Thread Metadata object.
    thread_metadata: dict | ThreadMetadata | None = field(  # noqa
        converter=ThreadMetadata, default=None  # noqa
    )  # noqa
    """Thread-specific fields not needed by other channels."""
    # TODO: implement Thread Member object.
    member: dict | ThreadMember | None = field(converter=ThreadMember, default=None)  # noqa
    """
    The thread member representation of the user if they have joined the thread.

    This is only included on certain api endpoints.
    """
    default_auto_archive_duration: int | None = field(default=None)
    """
    The default archive duration for threads in minutes.

    Can be set to `60`, `1440`, `4320`, `10080`.
    """
    permissions: str | None = field(default=None)
    """
    The computed permissions for the invoking user in the channel, including any overwrites.

    Only included when part of the resolved data received on a slash command interaction.
    """
    flags: int | ChannelFlags | None = field(converter=ChannelFlags, default=None)
    """Channel flags combined as a bitfield."""


@define(kw_only=True)
class StageChannel(Partial, Object):
    """
    Represents a stage channel from Discord.

    ---

    This is an event-specific dataclass that is not passed by
    Discord's Gateway. This will only contain the data relevant
    to a stage channel -- and just the `id` of it. `type` is
    dropped as this is already given for the event handler via.
    specification in the typehinting.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    bitrate : `int`, optional
        The bitrate of the voice channel.
    icon : `str`, optional
        The hash for the channel's icon.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    rtc_region : `str`, optional
        The channel's voice region ID if present, set to automatic when left as `None`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel."""
    type: int | ChannelType = field(converter=ChannelType)
    """The type of the channel."""
    guild_id: str | Snowflake | None = field(default=None, converter=Snowflake)
    """
    The ID of the guild.

    This is nullable due to some Gateway events
    lacking the data for the ID.
    """
    position: int | None = field(default=None)
    """Sorted position of the channel."""
    name: str | None = field(default=None)
    """
    The name of the channel.

    A channel name is in-between 1-100 characters.
    """
    bitrate: int | None = field(default=None)
    """The bitrate of the voice channel."""
    icon: str | None = field(default=None)
    """The hash for the channel's icon."""
    parent_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the parent of the channel

    Represents the parent category for regular channels and the parent channel for threads.
    """
    permissions: str | None = field(default=None)
    """
    The computed permissions for the invoking user in the channel, including any overwrites.

    Only included when part of the resolved data received on a slash command interaction.
    """
    flags: int | ChannelFlags | None = field(converter=ChannelFlags, default=None)
    """Channel flags combined as a bitfield."""


@define(kw_only=True)
class ThreadChannel(Partial, Object):
    """
    Represents a thread channel from Discord.

    ---

    This is an event-specific dataclass that is not passed by
    Discord's Gateway. This will only contain the data relevant
    to a thread channel -- and just the `id` of it. `type` is
    dropped as this is already given for the event handler via.
    specification in the typehinting.

    ---

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel."""
    guild_id: str | Snowflake | None = field(default=None, converter=Snowflake)
    """
    The ID of the guild.

    This is nullable due to some Gateway events
    lacking the data for the ID.
    """
    name: str | None = field(default=None)
    """
    The name of the channel.

    A channel name is in-between 1-100 characters.
    """
    last_message_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the last message sent in this channel.

    Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    """
    rate_limit_per_user: int | None = field(default=None)
    """
    Amount of seconds a user has to wait before sending another message

    Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    """
    icon: str | None = field(default=None)
    """The hash for the channel's icon."""
    owner_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the creator of the group dm or thread."""
    parent_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the parent of the channel

    Represents the parent category for regular channels and the parent channel for threads.
    """
    """The video quality mode of the voice channel."""
    message_count: int | None = field(default=None)
    """"
    The approximated amount of messages in a thread.

    Stops counting at `50`.
    """
    member_count: int | None = field(default=None)
    """
    The approximated amount of users in a thread.

    Stops counting at `50`.
    """
    # TODO: implement Thread Metadata object.
    thread_metadata: dict | ThreadMetadata | None = field(  # noqa
        converter=ThreadMetadata, default=None  # noqa
    )  # noqa
    """Thread-specific fields not needed by other channels."""
    # TODO: implement Thread Member object.
    member: dict | ThreadMember | None = field(converter=ThreadMember, default=None)  # noqa
    """
    The thread member representation of the user if they have joined the thread.

    This is only included on certain api endpoints.
    """
    default_auto_archive_duration: int | None = field(default=None)
    """
    The default archive duration for threads in minutes.

    Can be set to `60`, `1440`, `4320`, `10080`.
    """
    permissions: str | None = field(default=None)
    """
    The computed permissions for the invoking user in the channel, including any overwrites.

    Only included when part of the resolved data received on a slash command interaction.
    """
    flags: int | ChannelFlags | None = field(converter=ChannelFlags, default=None)
    """Channel flags combined as a bitfield."""


@define(kw_only=True)
class Channel(Object, Respondable):
    """
    Represents a channel from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the channel.
    type : `ChannelType`
        The type of the channel.
    guild_id : `Snowflake`, optional
        The ID of the guild.
        This is nullable due to some Gateway events
        lacking the data for the ID.
    position : `int`, optional
        Sorted position of the channel.
    permission_overwrites : `list[Overwrite]`, optional
        Explicit permission overwrites for members and roles.
    name : `str`, optional
        The name of the channel.

        A channel name is in-between 1-100 characters.
    topic : `str`, optional
        The topic of the channel.

        A channel topic is in-between 1-1024 characters.
    nsfw : `bool`, optional
        Whether or not the channel is NSFW. Defaults to `False`.
    last_message_id : `Snowflake`, optional
        The ID of the last message sent in this channel

        Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    bitrate : `int`, optional
        The bitrate of the voice channel.
    user_limit : `int`, optional
        The user limit of the voice channel.
    rate_limit_per_user : `int`, optional
        Amount of seconds a user has to wait before sending another message

        Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    recipients : `list[User]`, optional
        The recipients of the dm.
    icon : `str`, optional
        The hash for the channel's icon.
    owner_id : `Snowflake`, optional
        The ID of the creator of the group dm or thread.
    application_id : `Snowflake`, optional
        The ID of the application that created the dm if it is bot-created.
    parent_id : `Snowflake`, optional
        The ID of the parent of the channel

        Represents the parent category for regular channels and the parent channel for threads.
    last_pin_timestamp : `str`, optional
        The time when the last message was pinned.
    rtc_region : `str`, optional
        The channel's voice region ID if present, set to automatic when left as `None`..
    video_quality_mode : `VideoQualityMode`, optional
        The video quality mode of the voice channel.
    message_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    member_count : `int`, optional
        An approximate count of messages in a thread.

        Stops counting at `50`.
    thread_metadata : `ThreadMetadata`, optional
        Thread-specific fields not needed by other channels.
    member : `ThreadMember`, optional
        Thread member object for the current user if they have joined the thread

        This is only included on certain api endpoints.
    default_auto_archive_duration : `int`, optional
        The default archive duration for threads in minutes.

        Can be set to `60`, `1440`, `4320`, `10080`.
    permissions : `str`, optional
        Computed permissions for the invoking user in the channel, including overwrites.

        Only included when part of the resolved data received on a slash command interaction.
    flags : `ChannelFlags`, optional
        Channel flags combined as a bitfield.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel."""
    type: int | ChannelType = field(converter=ChannelType)
    """The type of the channel."""
    guild_id: str | Snowflake | None = field(default=None, converter=Snowflake)
    """
    The ID of the guild.

    This is nullable due to some Gateway events
    lacking the data for the ID.
    """
    position: int | None = field(default=None)
    """Sorted position of the channel."""
    permission_overwrites: list[dict] | list[Overwrite] | None = field(default=None)
    """Explicit permission overwrites for members and roles."""
    name: str | None = field(default=None)
    """
    The name of the channel.

    A channel name is in-between 1-100 characters.
    """
    topic: str | None = field(default=None)
    """
    The topic of the channel.

    A channel topic is in-between 1-1024 characters.
    """
    nsfw: bool | None = field(default=False)
    """Whether or not the channel is NSFW. Defaults to `False`."""
    last_message_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the last message sent in this channel.

    Can also be a thread if the channel is a forum. May not be an existing or valid message or thread.
    """
    bitrate: int | None = field(default=None)
    """The bitrate of the voice channel."""
    user_limit: int | None = field(default=None)
    """The user limit of the voice channel."""
    rate_limit_per_user: int | None = field(default=None)
    """
    Amount of seconds a user has to wait before sending another message

    Can be a number up to 21600. Bots, as well as users with the permission manage_messages or manage_channel, are unaffected.
    """
    recipients: list[dict] | list[User] | None = field(default=None)
    """The recipients of the dm."""
    icon: str | None = field(default=None)
    """The hash for the channel's icon."""
    owner_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the creator of the group dm or thread."""
    application_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """The ID of the application that created the dm if it is bot-created."""
    parent_id: str | Snowflake | None = field(converter=Snowflake, default=None)
    """
    The ID of the parent of the channel

    Represents the parent category for regular channels and the parent channel for threads.
    """
    last_pin_timestamp: str | None = field(default=None)
    """The time when the last message was pinned."""
    rtc_region: str | None = field(default=None)
    """The channel's voice region ID if present, set to automatic when left as `None`."""
    video_quality_mode: int | VideoQualityMode | None = field(
        converter=VideoQualityMode, default=None
    )
    """The video quality mode of the voice channel."""
    message_count: int | None = field(default=None)
    """"
    The approximated amount of messages in a thread.

    Stops counting at `50`.
    """
    member_count: int | None = field(default=None)
    """
    The approximated amount of users in a thread.

    Stops counting at `50`.
    """
    # TODO: implement Thread Metadata object.
    thread_metadata: dict | ThreadMetadata | None = field(  # noqa
        converter=ThreadMetadata, default=None  # noqa
    )  # noqa
    """Thread-specific fields not needed by other channels."""
    # TODO: implement Thread Member object.
    member: dict | ThreadMember | None = field(converter=ThreadMember, default=None)  # noqa
    """
    The thread member representation of the user if they have joined the thread.

    This is only included on certain api endpoints.
    """
    default_auto_archive_duration: int | None = field(default=None)
    """
    The default archive duration for threads in minutes.

    Can be set to `60`, `1440`, `4320`, `10080`.
    """
    permissions: str | None = field(default=None)
    """
    The computed permissions for the invoking user in the channel, including any overwrites.

    Only included when part of the resolved data received on a slash command interaction.
    """
    flags: int | ChannelFlags | None = field(converter=ChannelFlags, default=None)
    """Channel flags combined as a bitfield."""

    async def send(self, bot: Bot, content: str | None = None) -> "Message":  # noqa
        resp = await super().send(
            bot,
            f"/channels/{self.id}/messages",
            content=content,
        )
        # TODO: complete everything when everything is implemented, this is an example for now
        return Message(**resp)


@define(kw_only=True)
class _EmbedMedia:
    """
    Represents an embed thumbnail, video or image from Discord.

    Attributes
    ----------
    url : `str`, optional
        The url to the source of the media.

        The url is only needed for non-video contents.
    proxied_url : `str`, optional
        A proxied url of the media.
    height : `int`, optional
        The height of the media.
    width : `int`, optional
        The width of the media.
    """

    url: str | None = field(default=None)
    """
    The url to the source of the media.

    The url is only needed for non-video contents.
    """
    proxy_url: str | None = field(default=None)
    """A proxied url of the media."""
    height: int | None = field(default=None)
    """The height of the media."""
    width: int | None = field(default=None)
    """The width of the media."""


@define(kw_only=True)
class _EmbedProvider:
    """
    Represents the provider of an embed from Discord.

    Attributes
    ----------
    name : `str`, optional
        The of the provider.
    url : `str`, optional
        The url of the provider.
    """

    name: str | None = field(default=None)
    """The of the provider."""
    url: str | None = field(default=None)
    """The url of the provider."""


@define(kw_only=True)
class _EmbedAuthor:
    """
    Represents the author of an embed from Discord.

    Attributes
    ----------
    name : `str`, optional
        The name of the author.
    url : `str`, optional
        The URL of the author.
    icon_url : `str`, optional
        The URL of author's icon.
    proxy_icon_url : `str`, optional
        The proxied URL of the author's icon.
    """

    name: str | None = field(default=None)
    """The name of the author."""
    url: str | None = field(default=None)
    """The URL of the author."""
    icon_url: str | None = field(default=None)
    """The URL of author's icon."""
    proxy_icon_url: str | None = field(default=None)
    """The proxied URL of the author's icon."""


@define(kw_only=True)
class _EmbedFooter:
    """
    Represents the footer of an embed from Discord.

    Attributes
    ----------
    text : `str`, optional
        The footer text of an embed.
    icon_url : `str`, optional
        The URL of the footer icon.

        Only supports http(s) and attachments.
    proxy_icon_url : `str`, optional
        A proxied URL of the footer icon.
    """

    text: str | None = field(default=None)
    """The footer text of an embed."""
    icon_url: str | None = field(default=None)
    """
    The URL of the footer icon.

    Only supports http(s) and attachments.
    """
    proxy_icon_url: str | None = field(default=None)
    """A proxied URL of the footer icon."""


@define(kw_only=True)
class _EmbedField:
    """
    Represents a field of an embed from Discord.

    Attributes
    ----------
    name : `str`
        The name of the embed field.
    value : `str`
        The value of the embed field.
    inline : `bool`
        Whether or not the embed field is placed inline with other embed fields.
    """

    name: str = field()
    """The name of the embed field."""
    value: str = field()
    """The value of the embed field."""
    inline: bool | None = field(default=None)
    """Whether or not the embed field is placed inline with other embed fields."""


@define(kw_only=True, frozen=True)
class Embed:
    """
    Represnts a message embed from Discord.

    Attributes
    ----------
    title : `str`, optional
        The title of the embed.
    type : `str`, optional
        The type of the embed.

        This will always be `"rich"` for webhook embeds.
    description : `str`, optional
        The description of the embed.
    url : `str`, optional
        The url of the embed. This makes the title of the event a hyperlink.
    timestamp : `str`, optional
        The timestamp of the embed. This appears at the bottom.
    color : `Color`, optional
        The color of the embed.
    footer : `_EmbedFooter`, optional
        The footer of the embed.
    image : `_EmbedMedia`, optional
        The image of the embed. This is a large image displayed under the title.
    thumbnail : `_EmbedMedia`, optional
        The thumbnail of the embed. This is displayed as a small image on the embed.
    video : `_EmbedMedia`, optional
        The video of the embed.

        This cannot be used for embeds from bots.
    provider : `_EmbedMedia`, optional
        The provider of the embed.

        This cannot be used for embeds from bots.
    author : `_EmbedAuthor`, optional
        The author of the embed.
    fields : `list[_EmbedField]`, optional
        The fields of the embed.
    """

    title: str | None = field(default=None)
    """The title of the embed."""
    type: str | None = field(default=None)
    """
    The type of the embed.

    This will always be `"rich"` for webhook embeds.
    """
    description: str | None = field(default=None)
    """The description of the embed."""
    url: str | None = field(default=None)
    """The url of the embed. This makes the title of the event a hyperlink."""
    timestamp: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """The timestamp of the embed. This appears at the bottom."""
    # TODO: implement Color abc
    # color: int | Color | None = field(converter=optional_c(Color), default=None)
    """The color of the embed."""
    footer: dict | _EmbedFooter | None = field(converter=optional_c(_EmbedFooter), default=None)
    """The footer of the embed."""
    image: dict | _EmbedMedia | None = field(converter=optional_c(_EmbedMedia), default=None)
    """The image of the embed. This is a large image displayed under the title."""
    thumbnail: dict | _EmbedMedia | None = field(converter=optional_c(_EmbedMedia), default=None)
    """The thumbnail of the embed. This is displayed as a small image on the embed."""
    video: dict | _EmbedMedia | None = field(converter=optional_c(_EmbedMedia), default=None)
    """
    The video of the embed.

    This cannot be used for embeds from bots.
    """
    provider: dict | _EmbedProvider | None = field(
        converter=optional_c(_EmbedProvider), default=None
    )
    """
    The provider of the embed.

    This cannot be used for embeds from bots.
    """
    author: dict | _EmbedAuthor | None = field(converter=optional_c(_EmbedAuthor), default=None)
    """The author of the embed."""
    fields: list[dict] | list[_EmbedField] | None = field(
        converter=optional_c(list_c(_EmbedField), default=None)
    )
    """The fields of the embed."""


@define(kw_only=True)
class Attachment(Object):
    """
    Represents an attachment from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the attachment.
    filename : `str`, optional
        The name of the attached file.

        Always provided by Discord. Optional for bot use.
    description : `str`, optional
        The description of the attached file.
    content_type : `str`, optional
        The type of media attached.
    size : `int`, optional
        The size of the attached file in bytes.

        Always provided by Discord. Optional for bot use.
    url : `str`, optional
        The source URL of the file.

        Always provided by Discord. Optional for bot use.
    proxy_url : `str`, optional
        A proxied URL of the file.

        Always provided by Discord. Optional for bot use.
    height : `int`, optional
        The height of the file.

        This is only used for images.
    width : `int`, optional
        The height of the file.

        This is only used for images.
    ephemeral : `bool`, optional
        Whether or not the attachment is ephemeral.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the attachment."""
    filename: str | None = field(default=None)
    """
    The name of the attached file.

    Always provided by Discord. Optional for bot use.
    """
    description: str | None = field(default=None)
    """The description of the attached file."""
    content_type: str | None = field(default=None)
    """The type of media attached."""
    size: int | None = field(default=None)
    """
    The size of the attached file in bytes.

    Always provided by Discord. Optional for bot use.
    """
    url: str | None = field(default=None)
    """
    The source URL of the file.

    Always provided by Discord. Optional for bot use.
    """
    proxy_url: str | None = field(default=None)
    """
    A proxied URL of the file.

    Always provided by Discord. Optional for bot use.
    """
    height: int | None = field(default=None)
    """
    The height of the file.

    This is only used for images.
    """
    width: int | None = field(default=None)
    """
    The height of the file.

    This is only used for images.
    """
    ephemeral: bool | None = field(default=None)
    """Whether or not the attachment is ephemeral."""


@define(kw_only=True)
class Message(Object):
    """
    Represents a message from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the message.
    channel_id : `Snowflake`
        The ID of the channel containing the message.
    author : `User`
        The author of the message.
    content : `str`
        The content of the message.
    timestamp : `datetime`
        When the message was sent.
    edited_timestamp : `datetime`, optional
        When the message was last edited.
    tts : `bool`
        Whether or not this was a Text to Speech message.
    mention_everyone : `bool`
        Whether or not this message mentions everyone.
    mentions : `list[User]`
        Users specifically mentioned in this message.
    mention_roles : `list[Role]`
        Roles specifically mentioned in this message.
    mention_channels : `list[ChannelMention]`, optional
        Channels specifically mentioned in this message.
    attachments : `list[Attachment]`, optional
        The attachments of the message.
    embeds : `list[Embed]`
        The embeds of the message.
    reactions : `list[Reaction]`, optional
        The reactions to the message.
    nonce : `int | str`, optional
        Used for validating whether a message has been sent.
    pinned : `bool`
        Whether or not a message is pinned.
    webhook_id : `Snowflake`, optional
        The ID of the message's webhook, if the message was generated by a webhook.
    type : `MessageType`
        The type of the message.
    activity : `MessageActivity`, optional
        The activity of the message.

        Sent with Rich Presence related chat embeds.
    application : `Application`, optional
        The application of the message's activity.

        Sent with Rich Presence related chat embeds.
    application_id : `Snowflake`, optional
        The ID of the application.

        Sent if the message is an interactions or an application-owned webhook.
    message_reference : `MessageReference`, optional
        The source of a crosspost or the source of activity related to channel follows.
    flags : `MessageFlags`, optional
        Bitwise values representing a channel's flags.
    referenced_message : `Message`, optional
        The message associated with a message_reference.
    interaction : `Interaction`, optional
        The message's interaction if it is a response to an interaction.
    thread : `ThreadChannel`, optional
        The thread of the message, if it is the message that started a thread.
    components : `list[Component]`, optional
        The components on a message.
    sticker_items : `list[StickerItem]`, optional
        The items used to begin rendering the message's stickers.
    stickers : `list[Sticker]`, optional
        The stickers of a message.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the message."""
    channel_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the channel containing the message."""
    author: dict | User = field(converter=User)
    """The author of the message."""
    content: str = field()
    """The content of the message."""
    timestamp: str | datetime = field(converter=datetime.fromisoformat)
    """When the message was sent."""
    edited_timestamp: str | datetime | None = field(
        converter=optional_c(datetime.fromisoformat), default=None
    )
    """When the message was last edited."""
    tts: bool = field()
    """Whether or not this was a Text to Speech message."""
    mention_everyone: bool = field()
    """Whether or not this message mentions everyone."""
    mentions: list[dict] | list[User] = field(converter=list_c(User))
    """Users specifically mentioned in this message."""
    mention_roles: list[dict] | list[Role] = field(converter=list_c(Role))
    """Roles specifically mentioned in this message."""
    mention_channels: list[dict] | list[ChannelMention] | None = field(
        converter=optional_c(list_c(ChannelMention)), default=None
    )
    """Channels specifically mentioned in this message."""
    attachments: list[dict] | list[Attachment] = field(converter=list_c(Attachment))
    """The attachments of the message."""
    embeds: list[dict] | list[Embed] = field(converter=list_c(Embed))
    """The embeds of the message."""
    reactions: list[dict] | list[Reaction] | None = field(
        converter=optional_c(list_c(Reaction)), default=None
    )
    """The reactions to the message."""
    nonce: int | str | None = field(default=None)
    """Used for validating whether a message has been sent."""
    pinned: bool = field()
    """Whether or not a message is pinned."""
    webhook_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the message's webhook, if the message was generated by a webhook."""
    type: int | MessageType = field(converter=MessageType)
    """The type of the message."""
    activity: dict | MessageActivity | None = field(
        converter=optional_c(MessageActivity), default=None
    )
    """
    The activity of the message.

    Sent with Rich Presence related chat embeds.
    """
    application: dict | Application | None = field(converter=optional_c(Application), default=None)
    """
    The application of the message's activity.

    Sent with Rich Presence related chat embeds.
    """
    application_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """
    The ID of the application.

    Sent if the message is an interactions or an application-owned webhook.
    """
    message_reference: dict | MessageReference | None = field(
        converter=optional_c(MessageReference), default=None
    )
    """The source of a crosspost or the source of activity related to channel follows."""
    flags: int | MessageFlags | None = field(converter=optional_c(MessageFlags), default=None)
    """Bitwise values representing a channel's flags."""
    referenced_message: dict | "Message" | None = field(
        default=None
    )  # TODO: Implement recursive converters
    """The message associated with a message_reference."""
    # TODO: Implement Interaction object.
    # interaction: dict | Interaction | None = field(converter=optional_c(Interaction), default=None)
    """# The message's interaction if it is a response to an interaction."""
    thread: dict | ThreadChannel | None = field(converter=optional_c(ThreadChannel), default=None)
    """# The thread of the message, if it is the message that started a thread."""
    # TODO: Implement Component object.
    # components: list[dict] | list[Component] | None = field(converter=optional_c(list_c(Component)), default=None)
    # """The components on a message."""
    sticker_items: list[dict] | list[StickerItem] | None = field(converter=optional_c(list_c(StickerItem)), default=None)
    """The items used to begin rendering the message's stickers."""
    stickers: list[dict] | list[Sticker] | None = field(converter=optional_c(list_c(Sticker)), default=None)
    """The stickers of a message."""



from ...const import MISSING


class _MessageEvents:
    """
    Acts as a lookup table for Gateway events relating to
    messages from Discord.
    """

    @classmethod
    def lookup(cls, name: str, data: dict) -> dict | MISSING:
        match name:
            case "MESSAGE_CREATE":
                ...
            case "MESSAGE_UPDATE":
                ...
            case "MESSAGE_DELETE":
                ...
            case "MESSAGE_DELETE_BULK":
                ...
            case "MESSAGE_REACTION_ADD":
                ...
            case "MESSAGE_REACTION_REMOVE":
                ...
            case "MESSAGE_REACTION_REMOVE_ALL":
                ...
            case "MESSAGE_REACTION_REMOVE_EMOJI":
                ...
            case _:
                ...



from attrs import define, field
from enum import IntFlag

from ...utils.converters import optional_c

from .abc import Object, Partial, Snowflake

__all__ = ("PartialApplication", "Application", "ApplicationFlags", "InstallParams")


class ApplicationFlags(IntFlag):
    """
    Application flags are a set of bitwise values that represent the public flags
    of an application.
    """

    GATEWAY_PRESENCE = 1 << 12
    """The intent required for bots in 100 or more servers in order to receive presence_update events."""
    GATEWAY_PRESENCE_LIMITED = 1 << 13
    """The intent required for bots in under 100 servers in order to receive presence_update events, can be found in bot settings."""
    GATEWAY_GUILD_MEMBERS = 1 << 14
    """The intent required for bots in 100 or more servers in order to receive member-related events."""
    GATEWAY_GUILD_MEMBERS_LIMITED = 1 << 15
    """The intent required for bots in under 100 servers in order to receive member-related events, can be found in bot settings."""
    VERIFICATION_PENDING_GUILD_LIMIT = 1 << 16
    """Indicates unusual growth of an app that prevents verification."""
    EMBEDDED = 1 << 17
    """Indicates if an app is embedded within the Discord client (currently unavailable publicly)."""
    GATEWAY_MESSAGE_CONTENT = 1 << 18
    """The intent required for bots in 100 or more servers in order to receive message content"""
    GATEWAY_MESSAGE_CONTENT_LIMITED = 1 << 19
    """Intent required for bots in under 100 servers in order to receive message content, can be found in bot settings."""


@define(kw_only=True)
class InstallParams(Object):
    """
    Represents the install parameters of an application from Discord.

    Attributes
    ----------
    scopes : `list[str]`
        The scopes the application needs to join a server.
    permissions : `str`
        The permissions the bot requests be in the bot role.
    """

    scopes: list[str] = field()
    """The scopes the application needs to join a server."""
    permissions: str = field()
    """The permissions the bot requests be in the bot role."""


@define()
class PartialApplication(Partial):
    """
    Represents a partial application from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the application.
    flags : `ApplicationFlags`
        The public flags of the application.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the application."""
    flags: int | ApplicationFlags = field(converter=ApplicationFlags, default=0)
    """The public flags of the application. Defaults to `0`."""


@define(kw_only=True)
class Application(Object):
    """
    Represents an application from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the application.
    name : `str`
        The name of the application.
    icon : `str`
        The hash for the application's icon.

        This hash is pre-determined by the API and does not reflect
        the URL path.
    description : `str`
        The description of the application.
    rpc_origins : `list[str]`, optional
        A list of rpc origin urls, if rpc is enabled.
    bot_public : `bool`
        False if only application owner can join the application's bot
        to guilds.
    bot_require_code_grant : `bool`
        True if the application's bot has the oauth2 code grant flow enabled.
    terms_of_service_url : `str`, optional
        The url for the application's terms of service.
    privacy_policy_url : `str`, optional
        The url for the application's privacy policy.
    summary : `str`
        **Deprecated**. This is an empty string that will be removed in v11.
        Defaults to an empty string.
    verify_key : `str`
        The hex encoded key for verification in interactions and the
        gamesdk's getticket.
    guild_id : `Snowflake`, optional
        The ID of the guild if the application is a game sold on Discord.
    primary_sku_id : `Snowflake`, optional
        The ID of the "game sku" if it exists and the application is a
        game sold on Discord.
    slug : `str`, optional
        The url slug that links to the application's store page if it is a
        game sold on Discord.
    cover_image : `str`, optional
        The hash for the default rich presence invite cover.

        This hash is pre-determined by the API and does not reflect
        the URL path.
    flags : `ApplicationFlags`
        The public flags of the application.
    tags : `list[str]`, optional
        A maximum of 5 tags describing the content and functionality of the
        application.
    install_params : `InstallParams`, optional
        Settings for the application's default in-app authorization link.
    custom_install_url : `str`, optional
        The application's default custom authorization link.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the application."""
    name: str = field()
    """The name of the application."""
    # TODO: consider making icon hash an Image object
    icon: str = field()
    """
    The hash for the application's icon.

    This hash is pre-determined by the API and does not reflect
    the URL path.
    """
    description: str = field()
    """The description of the application."""
    rpc_origins: list[str] | None = field(default=None)
    """A list of rpc origin urls, if rpc is enabled."""
    bot_public: bool = field()
    """False if only application owner can join the application's bot to guilds."""
    bot_require_code_grant: bool = field()
    """True if the application's bot has the oauth2 code grant flow enabled."""
    terms_of_service_url: str | None = field(default=None)
    """The url for the application's terms of service."""
    privacy_policy_url: str | None = field(default=None)
    """The url for the application's privacy policy."""
    # TODO: implement User object
    # owner: dict | User | None = field(converter=User, default=None)  # noqa
    # """A partial user object representing the application's owner."""
    summary: str = field()
    """**Deprecated**. This is an empty string that will be removed in v11. Defaults to `""`"""
    verify_key: str = field()
    """The hex encoded key for verification in interactions and the gamesdk's getticket"""
    # TODO: implement Team object
    # team: dict | Team = field(converter=Team)  # noqa
    # """A team object representing the team that the application belongs to."""
    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the guild if the application is a sold game."""
    primary_sku_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the "game sku" if it exists and the application is a game sold on Discord."""
    slug: str | None = field(default=None)
    """IThe url slug that links to the application's store page if it is a game sold on Discord."""
    cover_image: str | None = field(default=None)
    """
    The hash for the default rich presence invite cover.

    This hash is pre-determined by the API and does not reflect
    the URL path.
    """
    flags: int | ApplicationFlags = field(converter=ApplicationFlags)
    """The public flags of the application."""
    tags: list[str] | None = field(default=None)
    """A maximum of 5 tags describing the content and functionality of the application."""
    install_params: dict | InstallParams | None = field(
        converter=optional_c(InstallParams), default=None
    )
    """The settings for the application's default in-app authorization link."""
    custom_install_url: str | None = field(default=None)
    """The application's default custom authorization link."""



from .abc import Snowflake, Object
from ...utils.converters import optional_c

from attrs import define, field


__all__ = (
    "Role",
    "RoleTags",
)


@define(kw_only=True)
class RoleTags:
    """
    Represents the tags of a role from Discord.

    Attributes
    ----------
    bot_id : `Snowflake`, optional
        The ID of the bot this role belongs to.
    integration_id : `Snowflake`, optional
        The ID of the integration this role belongs to.
    premium_subscriber : `bool`
        Whether this is the guild's premium subscriber role.
    """

    bot_id: Snowflake | str | None = field(converter=optional_c(Snowflake), default=None)
    """The id of the bot this role belongs to."""
    integration_id: Snowflake | str | None = field(converter=optional_c(Snowflake), default=None)
    """The id of the integration this role belongs to."""
    premium_subscriber: bool = field(default=False)
    """Whether this is the guild's premium subscriber role."""


@define(kw_only=True)
class Role(Object):
    """
    Represents a role from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the role.
    name : `str`
        The name of the role.
    color : `int`
        The integer representation of hexadecimal color code of the role.
    hoist : `bool`
        If this role is pinned in the user listing.
    icon : `str`, optional
        The hash of the roles icon, if present.
    unicode_emoji : `str`, optional
        The emoji unicode of the role, if present.
    position : `int`
        The position of the role.
    permissions : `str`
        The role's permission bit set.
    managed : `bool`
        Whether this role is managed by an integration.
    tags : `RoleTags`, optional
        The tags this role has, if present.
    """

    id: Snowflake | str = field(converter=Snowflake)
    """The ID of the role."""
    name: str = field()
    """The name of the role."""
    color: int = field()
    """The integer representation of hexadecimal color code of the role."""
    hoist: bool = field()
    """If this role is pinned in the user listing."""
    icon: str | None = field(default=None)
    """The role icon hash, if present."""
    unicode_emoji: str | None = field(default=None)
    """The emoji unicode of the role, if present."""
    position: int = field()
    """The position of the role."""
    permissions: str = field()
    """The role's permission bit set."""
    managed: bool = field()
    """Whether this role is managed by an integration."""
    mentionable: bool = field()
    """Whether this role is mentionable."""
    tags: RoleTags | None = field(converter=optional_c(RoleTags), default=None)
    """The tags this role has, if present."""



from enum import IntEnum

from attrs import define, field

from .abc import Object, Snowflake

from ...utils.converters import list_c, optional_c

__all__ = ("ApplicationCommand", "ApplicationCommandOption", "ApplicationCommandOptionChoice")


@define(kw_only=True)
class ApplicationCommandOptionChoice:
    """
    Represents a choice in an option for an application command from Discord.

    Attributes
    ----------
    name : `str`
        The name of the application command option choice in-between 1-100 characters.
    name_localizations : `dict[str, str]`, optional
        The localised dictionary of names for the application command option choices, if present.
    value : `str`, `int`, optional
        The value of the application command option type. The maximum length is 100 characters if `STRING`.
    """

    name: str = field()
    """The name of the application command option choice in-between 1-100 characters."""
    name_localizations: dict[str, str] | None = field(default=None)
    """The localised dictionary of names for the application command option choices, if present."""
    value: str | int = field()
    """The value of the application command option type. The maximum length is 100 characters if `STRING`."""


class ApplicationCommandOptionType(IntEnum):
    """Represents the types of application command options from Discord."""

    SUB_COMMAND = 1
    """A subcommand or "nested command.\""""
    SUB_COMMAND_GROUP = 2
    """A subcommand group or "nest or nested commands.\""""
    STRING = 3
    """A string input."""
    INTEGER = 4
    """An integer input."""
    BOOLEAN = 5
    """A boolean input."""
    USER = 6
    """A user/guild member input. Guild member will take priority if user is in the server on execution."""
    CHANNEL = 7
    """A guild channel input."""
    ROLE = 8
    """A guild role input."""
    MENTIONABLE = 9
    """A mentionable resource input."""
    NUMBER = 10
    """A number input."""
    ATTACHMENT = 11
    """An attachment input for attachment resources."""


@define(kw_only=True)
class ApplicationCommandOption:
    """
    Represents an option in an application command from Discord.

    Attributes
    ----------
    type : `ApplicationCommandOptionType`
        The type of application command option.
    name : `str`
        The name of the option in-between 1-32 characters.
    name_localizations : `dict[str, str]`, optional
        The localised dictionary of names for the application command option, if present.
    description : `str`
        The description of the option in-between 1-100 characters.
    description_localizations : `dict[str, str]`, optional
        The localised dictionary of descriptions for the application command option, if present.
    required : `bool`, optional
        Whether the application command option is required to be entered or not.
    choices : `list[ApplicationCommandOptionChoice]`, optional
        Pre-filled choices of an application command option, if present.

        The choices must be from a `STRING`, `INTEGER` or `NUMBER` type.
        An application command option can have a maximum of 25 choices.
    options : `list[ApplicationCommandOption]`, optional
        The options of the application command, if present.
        Options are only present on `CHAT_INPUT` command types.

        These options will only show if this type is `SUB_COMMAND` or
        `SUB_COMMAND_GROUP`.
    min_value : `int`, optional
        The minimum value permitted for the application command option.
    max_value : `int`, optional
        The maximum value permitted for the application command option.
    min_length : `int`, optional
        The minimum length permitted for the application command option.
    max_length : `int`, optional
        The maximum length permitted for the application command option.
    autocomplete : `bool`, optional
        Whether the application command option is autocompleted or not.

        The option must be from a `STRING`, `INTEGER` or `NUMBER` type.
    """

    type: int | ApplicationCommandOptionType = field(
        converter=optional_c(ApplicationCommandOptionType), default=None
    )
    """The type of application command option."""
    name: str = field()
    """The name of the option in-between 1-32 characters."""
    name_localizations: dict[str, str] | None = field(default=None)
    """The localised dictionary of names for the application command option, if present."""
    description: str = field()
    """The description of the option in-between 1-100 characters."""
    description_localizations: dict[str, str] | None = field(default=None)
    """The localised dictionary of descriptions for the application command option, if present."""
    required: bool | None = field(default=None)
    """Whether the application command option is required to be entered or not."""
    choices: list[ApplicationCommandOptionChoice] | None = field(
        converter=optional_c(list_c(ApplicationCommandOptionChoice)), default=None
    )
    """
    Pre-filled selection choices of an application command option.

    The choices must be from a `STRING`, `INTEGER` or `NUMBER` type.
    An application command option can have a maximum of 25 choices.
    """
    options: list[dict] | list["ApplicationCommandOption"] | None = field(
        converter=optional_c(list_c("ApplicationCommandOption")), default=None
    )
    """
    The options of the application command, if present.
    Options are only present on `CHAT_INPUT` command types.

    These options will only show if this type is `SUB_COMMAND` or
    `SUB_COMMAND_GROUP`.
    """

    # TODO: implement a channel type integer enumerable.
    # channel_types: list[ChannelType] | None = field(default=None)
    # """The types of channels the option will filter to, if present."""

    min_value: int | None = field(default=None)
    """The minimum value permitted for the application command option."""
    max_value: int | None = field(default=None)
    """The maximum value permitted for the application command option."""
    min_length: int | None = field(default=None)
    """The minimum length permitted for the application command option. The minimum allowed is `0`."""
    max_length: int | None = field(default=None)
    """The maximum length permitted for the application command option. The maximum allowed is `1`."""
    autocomplete: bool | None = field(default=None)
    """
    Whether the application command option is autocompleted or not.

    The option must be from a `STRING`, `INTEGER` or `NUMBER` type.
    """


class ApplicationCommandType(IntEnum):
    """Represents the types of application commands from Discord."""

    CHAT_INPUT = 1
    """A chat-input command, or "slash command.\""""
    USER = 2
    """A user command, or "user context menu.\""""
    MESSAGE = 3
    """A message command, or "message context menu.\""""


@define(kw_only=True)
class ApplicationCommand(Object):
    """
    Represents an application command from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the application command.
    type : `ApplicationCommandType`
        The type of application command.
    application_id : `Snowflake`
        The ID of the application the command is under.
    guild_id : `Snowflake`, optional
        The ID of the guild the command is under, if present.
    name : `str`
        The name of the command in-between 1-32 characters.
    name_localizations : `dict[str, str]`, optional
        The localised dictionary of names for the application command, if present.
    description : `str`
        The description of the command in-between 1-100 characters.
        Descriptions are only present on `CHAT_INPUT` command types.
    description_localizations : `dict[str, str]`, optional
        The localised dictionary of descriptions for the application command, if present.
    options : `list[ApplicationCommandOption]`, optional
        The options of the application command, if present.
        A maximum of 25 options are allowed. Options are only
        present on `CHAT_INPUT` command types.
    default_member_permissions : `str`, optional
        The default permissions of the application command, if present.
    dm_permission : `bool`, optional
        Whether the application command is able to be ran in DMs or not.
    version : `Snowflake`, optional
        The internal version of application commands released. This auto-increments over time.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the application command."""
    type: int | ApplicationCommandType = field(converter=ApplicationCommandType)
    """The type of application command."""
    application_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the application the command is under."""
    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the guild the command is under, if present."""
    name: str = field()
    """The name of the command in-between 1-32 characters."""
    name_localizations: dict[str, str] | None = field(default=None)
    """The localised dictionary of names for the application command, if present."""
    description: str = field()
    """
    The description of the command in-between 1-100 characters.
    Descriptions are only present on `CHAT_INPUT` command types.
    """
    description_localizations: dict[str, str] | None = field(default=None)
    """The localised dictionary of descriptions for the application command, if present."""
    options: list[dict] | list[ApplicationCommandOption] | None = field(
        converter=optional_c(list_c(ApplicationCommandOption)), default=None
    )
    """
    The options of the application command, if present.
    A maximum of 25 options are allowed. Options are only
    present on `CHAT_INPUT` command types.
    """
    default_member_permissions: str | None = field(default=None)
    """The default permissions of the application command, if present."""
    dm_permission: bool | None = field(default=None)
    """Whether the application command is able to be ran in DMs or not."""
    version: str | Snowflake = field(converter=Snowflake)
    """The internal version of application commands released. This auto-increments over time."""



from enum import IntEnum
from attrs import define, field
from .abc import Object, Snowflake
from .user import User
from ...utils.converters import list_c, optional_c


class StickerType(IntEnum):
    """
    Represents the types of stickers from Discord.
    
    Constants
    ---------
    STANDARD
        An official sticker of a pack. 
    
        Only availible with Nitro unless
        part of a removed purchaseable pack.
    GUILD
        A sticker uploaded by a guild.
    """
    STANDARD = 1
    """
    An official sticker of a pack. 
    
    Only availible with Nitro unless
    part of a removed purchaseable pack.
    """
    GUILD = 2
    """A sticker uploaded by a guild."""


class StickerFormatType(IntEnum):
    """
    Represents the types of sticker formats from Discord.
    
    Constants
    ---------
    PNG
        A sticker formatted with the `.png` format.
    APNG
        A sticker formatted with the `.apng` format.
    LOTTIE
        A sticker formatted with the `.lottie` format.
    """
    PNG = 1
    """A sticker formatted with the `.png` format."""
    APNG = 2
    """A sticker formatted with the `.apng` format."""
    LOTTIE = 3
    """A sticker formatted with the `.lottie` format."""


@define(kw_only=True)
class Sticker(Object):
    """
    Represents a sticker from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the sticker.
    pack_id : `Snowflake`, optional
        The ID of the pack containing the sticker.

        Only standard stickers owned by Discord
        can be part of a sticker pack.
    name : `str`
        The name of the sticker.
    description : `str`, optional
        The description of the sticker.
    tags : `str`
        The tags or emoji equivelent of a sticker.

        If the sticker is a standard sticker, this
        field will be a comma seperated list of 
        keywords. If the sticker is guild owned, it
        will be the name of an emoji.
    asset : `str`
        An empty string previously for the sticker asset hash.
    type : `StickerType`
        The type of the sticker.
    format_type : `StickerFormatType`
        The formatting type that the sticker uses.
    available : `bool`, optinal
        Whether or not the sticker is availible.

        May be false if the guild that owns it
        has lost a level of boosting.
    guild_id : `Snowflake`, optional
        The ID of the guild that owns the sticker.

        Will only be set to None if the sticker is
        owned by Discord.
    user : `User`, optional
        The user that uploaded the sticker.

        Will only be set to None if the sticker is
        owned by Discord.
    sort_value : `int`, optional
        The sticker's order within its sticker pack.

        Will only be availible if the sticker is
        part of a pack and owned by Discord.
    """
    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the sticker."""
    pack_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """
    The ID of the pack containing the sticker.
    
    Only standard stickers owned by Discord
    can be part of a sticker pack.
    """
    name: str = field()
    """The name of the sticker."""
    description: str | None = field(default=None)
    """The description of the sticker."""
    tags: str = field()
    """
    The tags or emoji equivelent of a sticker.
    
    If the sticker is a standard sticker, this
    field will be a comma seperated list of 
    keywords. If the sticker is guild owned, it
    will be the name of an emoji.
    """
    asset: str = field(default="")
    """An empty string previously for the sticker asset hash."""
    type: int | StickerType = field(converter=StickerType)
    """The type of the sticker."""
    format_type: int | StickerFormatType = field(converter=StickerFormatType)
    """The formatting type that the sticker uses."""
    available: bool = field(default=True)
    """
    Whether or not the sticker is availible.
    
    May be false if the guild that owns it
    has lost a level of boosting.
    """
    guild_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """
    The ID of the guild that owns the sticker.
    
    Will only be set to None if the sticker is
    owned by Discord.
    """
    user: dict | User | None = field(converter=optional_c(User), default=None)
    """
    The user that uploaded the sticker.
    
    Will only be set to None if the sticker is
    owned by Discord.
    """
    sort_value: int | None = field(default=None)
    """
    The sticker's order within its sticker pack.

    Will only be availible if the sticker is
    part of a pack and owned by Discord.
    """
    @property
    async def clean_tags(self):
        """
        A clean list of tags, formatted from the `tags` attribute.
        
        Only availible on standard stickers.
        """
        if self.type == StickerType.STANDARD:
            return self.tags.split(", ")

@define(kw_only=True)
class StickerItem(Object):
    """
    Represents the bare information needed 
    to begin loading a sticker from Discord.
    
    Attributes
    ----------
    id : `Snowflake`
        The ID of the sticker being represented.
    name : `str`
        The name of the sticker being represented.
    format_type : `StickerFormatType`
        The type of formatting of the sticker being represented.
    """
    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the sticker being represented."""
    name: str = field()
    """The name of the sticker being represented."""
    format_type: int | StickerFormatType = field(converter=StickerFormatType)
    """The type of formatting of the sticker being represented."""


@define(kw_only=True)
class StickerPack(Object):
    """
    Represents a sticker pack from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID of the sticker pack.
    stickers : `list[Sticker]`
        The stickers within the sticker pack.
    name : `str
        The name of the sticker pack.
    sku_id : `Snowflake`
        The ID of the pack's SKU.
    cover_sticker_id : `Snowflake`, optional
        The ID of the sticker featured as the pack's icon.
    description : `str`
        The description of the sticker pack.
    banner_asset_id : `Snowflake`, optional
        The ID of the sticker packs's banner image.
    """
    id: str | Snowflake = field(converter=Snowflake)
    """The ID of the sticker pack."""
    stickers: list[dict] | list[Sticker] = field(converter=list_c(Sticker))
    """The stickers within the sticker pack."""
    name: str = field()
    """The name of the sticker pack."""
    sku_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the pack's SKU."""
    cover_sticker_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the sticker featured as the pack's icon."""
    description: str = field()
    """The description of the sticker pack."""
    banner_asset_id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """The ID of the sticker packs's banner image."""



from attrs import define, field
from .abc import Object, Snowflake
from .user import User
from .role import Role
from ...utils import optional_c, list_c


@define()
class Emoji(Object):
    """
    Represents an Emoji from Discord.

    Attributes
    ----------
    id : `Snowflake`, optional
        The ID of the Emoji.

        Only available on guild created emojis.
    name : `str`, optional
        The name of the emoji.

        This is a unicode emoji if the emoji
        is standard, otherwise it is the name
        of the emoji.
    user : `User`, optional
        The user that created the emoji.
    require_colons : `bool`, optional
        Whether or not the emoji needs to be wrapped in colons.
    managed : `bool`, optional
        Whether or not the emoji is managed.
    animated : `bool`, optional
        Whether or not the emoji is animated.
    available : `bool`
        Whether or not the emoji is available.

        Defaults to `False` if the server that owns
        the emoji lost a level of boosting.
    """

    id: str | Snowflake | None = field(converter=optional_c(Snowflake), default=None)
    """
    The ID of the Emoji.

    Only available on guild created emojis.
    """
    name: str | None = field(default=None)
    """
    The name of the emoji.

    This is a unicode emoji if the emoji
    is standard, otherwise it is the name
    of the emoji.
    """
    # roles: list[dict] | list[Role] | None = field(converter=optional_c(list_c(Role)))
    # TODO: Implement Role object
    # """A list of roles allowed to use this emoji."""
    user: dict | User | None = field(converter=optional_c(User), default=None)
    """The user that created the emoji."""
    require_colons: bool | None = field(default=None)
    """Whether or not the emoji needs to be wrapped in colons."""
    managed: bool | None = field(default=None)
    """Whether or not the emoji is managed."""
    animated: bool | None = field(default=None)
    """Whether or not the emoji is animated."""
    available: bool = field(default=False)
    """
    Whether or not the emoji is available.

    Defaults to `False` if the server that owns
    the emoji lost a level of boosting.
    """



from logging import getLogger
from typing import Any, Callable, Coroutine, Optional, Protocol

from trio import run

from ..api import GatewayClient
from ..api.http import HTTPClient
from ..const import MISSING, NotNeeded
from .flags import Intents

logger = getLogger(__name__)


class BotProtocol(Protocol):
    def __init__(self, intents: Intents):
        ...

    def start(self, token: str):
        ...

    def close(self):
        ...

    async def restart(self):
        ...

    async def on(self, coro: Coroutine, name: NotNeeded[str] = MISSING) -> Callable[..., Any]:
        ...

    @property
    def latency(self) -> float:
        ...

    @property
    def offline(self) -> bool:
        ...


class Bot(BotProtocol):
    """
    Represents a bot's connection to Discord.

    Attributes
    ----------
    intents : `Intents`
        The bot's intents.
    _gateway : `GatewayClient`
        The bot's gateway connection.
    http : `HTTPClient`
        The bot's HTTP connection.
    _calls : `dict[str, list[typing.Coroutine]]`
        A set of callbacks registered by their name to their function.
        These are used to help dispatch Gateway events.
    """

    intents: Intents
    """The bot's intents."""
    _gateway: GatewayClient
    """The bot's gateway connection."""
    http: HTTPClient
    """The bot's HTTP connection."""
    _calls: dict[str, list[Coroutine]] = {}
    """
    A set of callbacks registered by their name to their function.
    These are used to help dispatch Gateway events.
    """

    def __init__(self, intents: Intents):
        self.intents = intents
        self._gateway = MISSING
        self.http = MISSING

    def start(self, token: str):
        """
        Starts a connection with Discord.

        Parameters
        ----------
        token : `str`
            The token of the bot.
        """
        self.http = HTTPClient(token)
        run(self._connect, token)

    def close(self):
        """Closes the current connection with Discord."""
        self._gateway._stopped = True

    async def restart(self):
        """Restarts a connection with Discord."""
        await self._gateway.reconnect()

    async def _connect(self, token: str):
        """
        Connects to the Gateway and hooks into the manager.

        Parameters
        ----------
        token : `str`
            The token of the bot.
        """
        async with GatewayClient(token, self.intents) as self._gateway:
            await self._gateway._hook(self)

    def _register(self, coro: Coroutine, name: Optional[str] = None, event: Optional[bool] = True):
        """
        Registers a coroutine to be used as a callback.

        Parameters
        ----------
        coro : `typing.Coroutine`
            The coroutine associated with the event.
        name : `str`, optional
            The name associated with the event. Defaults to
            the name of the coroutine, prefixed with `on_`.
        event : `bool`, optional
            Whether the coroutine is a Gateway event or not.
            Defaults to `True`.
        """
        _name = (name if event else name) if name else coro.__name__

        logger.debug(f"Registering callback for {_name}.")
        call = self._calls.get(_name, [])
        call.append(coro)

        self._calls[_name] = call

    async def _trigger(self, name: str, *args):
        """
        Triggers a name registered for callbacks.

        Parameters
        ----------
        name : `str`
            The name associated with the callbacks.
        """
        for event in self._calls.get(name, []):
            await event(*args)

    def on(
        self, coro: NotNeeded[Coroutine] = MISSING, *, name: NotNeeded[str] = MISSING
    ) -> Callable[..., Any]:
        """
        Listens to events given from the Gateway.

        ---

        `@on` is a decorator that attaches onto an asynchronous
        function or method intended to receive dispatched Gateway events.

        ---

        Examples
        --------
        In retux, `event` is an argument containing the information relating
        to the event you're listening to. This argument is always required
        for an `@on` coroutine. `event` can be anything so as long as it's
        dispatched and subclassed as a `_Event`. In any case, however, you
        can always typehint it as what the event is intending to listen to.
        ```
        @bot.on
        async def ready(event):
            print(f"Hello, world! I am {event.user.username}. Running on API version v{event.version}.")
        ```

        If you'd like to still take control of the coroutine's name,
        however, you can simply pass the event name into the `@on`
        decorator.
        ```
        @bot.on("ready")
        async def func_name(event):
            ...
        ```

        `@on` empowers developers to also determine when a restart may
        be needed without having to create your own loop. Please note
        that this is only an example. retux automatically attempts reconnections
        if one is called for, including invalidated sessions.
        ```
        @bot.on
        async def reconnect(event: typing.Any):
            if bot.is_offline:
                await bot.restart()
        ```

        Sometimes, however, you may want to provide something a little bit
        more structural and organised than numerous functions laid out all
        over the place. Instead of resorting these into an extension,
        we have created what's known as "event classes" for this instead.
        ```py
        @bot.on
        class MessageEvent:
            async def create(message: retux.Message):
                ...
            async def remove(message, before, after):
                ...
        ```

        The nicest part of this is that you can create your own event handlers
        for attributes changing, done with `async def channel_id(value)`!

        Additionally, you can also provide converters for your message event,
        in case you do not feel like specifying it via. `create` and etc., such
        as `async def to_voice(channel: retux.VoiceChannel)`.

        If a bot developer so wishes to create their own functions inside of
        the class as well, it is highly recommended to dunder (add an `_` to
        the beginning of) their custom-defined methods. This poses no issue
        to retux's handling, but helps you as the bot developer keep up with
        what you're doing.

        Parameters
        ----------
        coro : `typing.Coroutine`, optional
            The coroutine to associate with the event. This is
            to be placed as a decorator on top of an asynchronous
            function.
            This is only "optional" when `name` has been specified.
        name : `str`, optional
            The name associated with the event. This defaults to the
            name of the coroutine, prefixed with `on_`.

        Returns
        -------
        `typing.Callable[..., typing.Any]`
            The coroutine associated with the event, with
            a callable pattern as to `coro`.
        """

        def decor(coro: Coroutine):
            self._register(coro, name=name if name is not MISSING else coro.__name__)
            return coro

        if coro is not MISSING:
            return decor(coro)

        return decor

    @property
    def latency(self) -> float:
        """
        The bot's latency from the Gateway.
        This is only calculated from heartbeats.
        """
        return self._gateway.latency

    @property
    def offline(self) -> bool:
        """
        Whether the bot is offline or not.
        May be useful for determining when to restart!
        """
        return bool(self._gateway._closed)



from .abc import Snowflake
from .user import User
from .guild import Guild

from attrs import define, field
from datetime import datetime

__all__ = ("GuildTemplate",)


@define()
class GuildTemplate:
    """
    Represents the template of a guild from Discord.

    Attributes
    ----------
    code : `str`
        The template code as unique ID.
    name : `str`
        The name of the template.
    description : `str`, optional
        The description of the template, if any.
    usage_count : `int`
        The amount of times this template has been used.
    creator_id : `Snowflake`
        The ID of the user who created the template.
    creator : `User`
        The user who created the template.
    created_at : `datetime`
        When this template was created.
    updated_at : `datetime`
        When this template was last synced to the source guild.
    source_guild_id : `Snowflake`
        The ID of the guild this template is based on.
    serialized_source_guild : `Guild`
        The guild snapshot this template contains.
    is_dirty : `bool`
        Whether the template has unsynced changes or not.
    """

    code: str = field()
    """The template code as unique ID."""
    name: str = field()
    """The name of the template."""
    description: str | None = field(default=None)
    """The description of the template, if any."""
    usage_count: int = field()
    """The amount of times this template has been used."""
    creator_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the user who created the template."""
    creator: dict | User = field(converter=User)
    """The user who created the template."""
    created_at: str | datetime = field(converter=datetime.fromisoformat)
    """When this template was created."""
    updated_at: str | datetime = field(converter=datetime.fromisoformat)
    """When this template was last synced to the source guild."""
    source_guild_id: str | Snowflake = field(converter=Snowflake)
    """The ID of the guild this template is based on."""
    serialized_source_guild: dict | Guild = field(converter=Guild)
    """The guild snapshot this template contains."""
    is_dirty: bool = field(default=False)
    """Whether the template has unsynced changes or not."""



from datetime import datetime
from enum import Enum
from typing import Union

from attrs import define, field

__all__ = ("Snowflake", "Image", "Partial", "Object", "Component")


@define(repr=False, eq=False)
class Snowflake:
    """
    Represents an unique identifier for a Discord resource.

    ---

    Discord utilizes Twitter's snowflake format for uniquely identifiable descriptors
    (IDs). These IDs are guaranteed to be unique across all of Discord, except in some
    unique scenarios in which child objects share their parent's ID.

    ---

    Attributes
    ----------
    _snowflake : `str`
        The internally stored snowflake. Snowflakes are always in string-form.

        This value should never need to be directly checked. Please use the
        representation of the class itself to do this for identity comparisons.

    Methods
    -------
    timestamp : `datetime.datetime`
        The timestamp of the snowflake as a UTC-native datetime.

        Timestamps are denoted as milliseconds since the Discord Epoch:
        the first second of 2015, or or `1420070400000`.
    worker_id : `int`
        The internal worker ID of the snowflake.
    process_id : `int`
        The internal process ID of the snowflake.
    increment : `int`
        The internal incrementation number of the snowflake.

        This value will only increment when a process has been
        generated on this snowflake, e.g. a resource.
    """

    _snowflake: str | int = field(converter=str)
    """
    The internally stored snowflake. Snowflakes are always in string-form.

    The snowflake may only be `None` in the event that a given
    field in a resource does not supply it. This should not be always
    taken for granted as having a value. Please use the representation
    of the class itself.
    """

    def __repr__(self) -> str | None:
        return self._snowflake

    def __eq__(self, other: Union[str, int, "Snowflake"]) -> bool:
        if type(other) == int:
            return bool(int(self._snowflake) == other)
        else:
            return bool(self._snowflake == str(other))

    @property
    def timestamp(self) -> datetime:
        """
        The timestamp of the snowflake as a UTC-native datetime.

        Timestamps are denoted as milliseconds since the Discord Epoch:
        the first second of 2015, or `1420070400000`.
        """
        retrieval: int | float = (int(self._snowflake) >> 22) + 1420070400000
        return datetime.utcfromtimestamp(retrieval)

    @property
    def worker_id(self) -> int:
        """The internal worker ID of the snowflake."""
        return (int(self._snowflake) & 0x3E0000) >> 17

    @property
    def process_id(self) -> int:
        """The internal process ID of the snowflake."""
        return (int(self._snowflake) & 0x1F000) >> 12

    @property
    def increment(self) -> int:
        """
        The internal incrementation number of the snowflake.

        This value will only increment when a process has been
        generated on this snowflake, e.g. a resource.
        """
        return int(self._snowflake) & 0xFFF


class CDNEndpoint(Enum):
    """
    Represents all of the different Discord
    CDN endpoints.

    Constants
    ---------
    CUSTOM_EMOJI
        The endpoint for custom emojis.
        
        The extra id is the emoji's ID.
    GUILD_ICON
        The endpoint for guild icons.
        
        The extra id is the guild's ID.
        Guild icons can be animated.
    GUILD_SPLASH
        The endpoint for guild splashes.
        
        The extra id is the guild's ID.
    GUILD_DISCOVERY_SPLASH
        The endpoint for guild discovery splashes.
        
        The extra id is the guild's ID.
    GUILD_BANNER
        The endpoint for guild banners.
        
        The extra id is the guild's ID.
    Guild banners can be animated.
    USER_BANNER
        The endpoint for user banners.
        
        The extra id is the user's ID.
    User banners can be animated.
    DEFAULT_USER_AVATAR
        The endpoint for the default avatars of users.
        
        The  extra id is the modulo 5 of
        the user's discriminator. The size param is 
        ignored for this endpoint.
    USER_AVATAR
        The endpoint for user avatars.
        
        The extra id is the user's ID.
        User avatars can be animated
    GUILD_MEMBER_AVATAR
        The endpoint for guild-specific member avatars.
        
        The extra ids are the guild's ID 
        and the user's ID. These can be animated.
    APPLICATION_ICON
        The endpoint for application icons.
        
        The extra id is the application's ID.
    APPLICATION_COVER
        The endpoint for application covers.
        
        The extra id is the application's ID.
    APPLICATION_ASSET
        The endpoint for application assets.
        
        The extra id is the applciation's ID.
    ACHIEVEMENT_ICON
        The endpoint for application icons.
        
        The extra id is the application's ID.
    STICKER_PACK_BANNER
        The endpoint for sticker pack banners.
        
        The extra id is the sticker pack
        banner asset's ID.
    TEAM_ICON
        The endpoint for team icons.
        
        The extra id is the team's ID.
    STICKER
        The endpoint for stickers.
        
        The extra id is the sticker's ID.
        The size parameter is ignored. Stickers can
        be animated.
    ROLE_ICON
        The endpoint for role icons.
        
        The extra id is the role's ID.
    GUILD_SCHEDULED_EVENT_COVER
        The endpoint for guild scheduled event covers.
        
        The extra id is the guild's ID.
    GUILD_MEMBER_BANNER
        The endpoint for guild member banners.
        
        The extra ids are the guild's ID
        and the user's ID.
    """
    CUSTOM_EMOJI = "emojis/{}"
    """
    The endpoint for custom emojis.
    
    The extra id is the emoji's ID.
    """
    GUILD_ICON = "icons/{}/{hash}"
    """
    The endpoint for guild icons.
    
    The extra id is the guild's ID.
    Guild icons can be animated.
    """
    GUILD_SPLASH = "splashes/{}/{hash}"
    """
    The endpoint for guild splashes.
    
    The extra id is the guild's ID.
    """
    GUILD_DISCOVERY_SPLASH = "discovery-splashes/{}/{hash}"
    """
    The endpoint for guild discovery splashes.
    
    The extra id is the guild's ID.
    """
    GUILD_BANNER = "banners/{}/{hash}"
    """
    The endpoint for guild banners.
    
    The extra id is the guild's ID.
    Guild banners can be animated.
    """
    USER_BANNER = "banners/{}/{hash}"
    """
    The endpoint for user banners.
    
    The extra id is the user's ID.
    User banners can be animated.
    """
    DEFAULT_USER_AVATAR = "embed/avatars/{}"
    """
    The endpoint for the default avatars of users.
    
    The  extra id is the modulo 5 of
    the user's discriminator. The size param is 
    ignored for this endpoint.
    """
    USER_AVATAR = "avatars/{}/{hash}"
    """
    The endpoint for user avatars.
    
    The extra id is the user's ID.
    User avatars can be animated
    """
    GUILD_MEMBER_AVATAR = "guilds/{}/users/{}/avatars/{hash}"
    """
    The endpoint for guild-specific member avatars.
    
    The extra ids are the guild's ID 
    and the user's ID. These can be animated.
    """
    APPLICATION_ICON = "app-icons/{}/{hash}"
    """
    The endpoint for application icons.
    
    The extra id is the application's ID.
    """
    APPLICATION_COVER = "app-icons/{}/{hash}"
    """
    The endpoint for application covers.
    
    The extra id is the application's ID.
    """
    APPLICATION_ASSET = "app-assets/{}/{hash}"
    """
    The endpoint for application assets.
    
    The extra id is the applciation's ID.
    """
    ACHIEVEMENT_ICON = "app-assests/{}/{hash}"
    """
    The endpoint for application icons.
    
    The extra id is the application's ID.
    """
    STICKER_PACK_BANNER = "app-assets/710982414301790216/store/{}"
    """
    The endpoint for sticker pack banners.
    
    The extra id is the sticker pack
    banner asset's ID.
    """
    TEAM_ICON = "team-icons/{}/{hash}"
    """
    The endpoint for team icons.
    
    The extra id is the team's ID.
    """
    STICKER = "stickers/{hash}"
    """
    The endpoint for stickers.
    
    The extra id is the sticker's ID.
    The size parameter is ignored. Stickers can
    be animated.
    """
    ROLE_ICON = "role-icons/{}/{hash}"
    """
    The endpoint for role icons.
    
    The extra id is the role's ID.
    """
    GUILD_SCHEDULED_EVENT_COVER = "guild-events/{}/{hash}"
    """
    The endpoint for guild scheduled event covers.
    
    The extra id is the guild's ID.
    """
    GUILD_MEMBER_BANNER = "guilds/{}/users/{}/banners/{hash}"
    """
    The endpoint for guild member banners.
    
    The extra ids are the guild's ID
    and the user's ID.
    """


@define(kw_only=True)
class Image:
    """
    Represents an image from Discord.

    Attributes
    ----------
    hash : `str`, optional
        The hash of the image.
    
    Methods
    -------
    animated : `bool`
        Whether or not the image is animated.
    path : `str`
        The path to the image in CDN.
    url : `str`
        The url to the image.
    """
    hash: str | None = field(default=None)
    """The hash of the image."""
    _endpoint: CDNEndpoint = field()
    """The endpoint of the image."""
    _ids: list | None = field(default=None)
    """
    The needed variables aside 
    from the hash for the endpoint.
    """

    def _c(endpoint: CDNEndpoint):
        """
        For internal use only.

        Builds an attrs converter for Images.
        A `__post_attrs_init__` must be used
        to add extra variables needed in the
        endpoint such as ids. This is only 
        meant for endpoints that need a hash.
        
        Please note that the naming of this method, `_c`
        does not reflect the same serialisation methods
        or strategies as `Serialisable._c()`.

        Parameters
        ----------
        endpoint : CDNEndpoint
            The endpoint of the image.
        
        Returns
        -------
        `function`
            A function that takes one argument
            (the hash). Compatible with attrs.
        """
        def inner(hash: str):
            return Image(hash=hash, endpoint=endpoint)
        return inner
    
    @property
    def animated(self) -> bool:
        """Whether or not the image is animated."""
        return self.hash and self.hash.startswith("a_")

    @property
    def path(self) -> str:
        """The path to the image in CDN."""
        if not self._ids:
            raise RuntimeError("Tried to access image endpoint without needed ids.")
        if "{hash}" in self._endpoint.value:
            return self._endpoint.value.format(*self._ids, hash=self.hash)
        else:
            return self._endpoint.value.format(*self._ids)

    @property
    def url(self) -> str:
        """The url to the image."""
        return f"https://cdn.discordapp.com/{self.path}"
    

@define()
class Partial:
    """
    Represents partial information to a resource from Discord.

    ---

    Sometimes, Discord will provide back to the client what is
    known as a "partial object." These objects are semantically
    categorised by their resource, but in cases do not carry
    the full set of information required for them. The `Partial`
    class lives to serve as a way to better typehint this incomplete
    data.
    """


@define(kw_only=True)
class Object:
    """
    Represents the base object form of a resource from Discord.

    Attributes
    ----------
    id : `Snowflake`
        The ID associated to the object.
    """

    id: str | Snowflake = field(converter=Snowflake)
    """The ID associated to the object."""


@define()
class Component:
    """
    Represents the base information of a component from Discord.

    ---

    `custom_id` is an attribute shared in every component,
    however, only `Button` makes them optional. A custom ID is
    a developer-defined ID in-between 1-100 characters.
    """

    custom_id: str | None = field(default=None)



from enum import Enum
from json import dumps, loads
from logging import INFO, getLogger
from sys import version_info
from typing import Protocol

from attrs import define, field
from httpx import AsyncClient, QueryParams, Response, __version__ as __http_version__
from trio import Event, sleep

from .error import HTTPException

from ..const import MISSING, NotNeeded, __api_url__, __repo_url__, __version__

logger = getLogger(__name__)

__all__ = ("_RouteMethod", "_Route", "_Limit", "HTTPClient")


class _RouteMethod(Enum):
    """Represents the types of route methods."""

    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    DELETE = "DELETE"
    PATCH = "PATCH"


@define(slots=False)
class _Route:
    """
    Represents a route path to an API endpoint.

    By standard nature, the route will only contain the relevant sending details,
    such as the method and path. However, in the event of a rate limit that applies
    to the route specifically, some fields will be populated such as `channel_id`
    and `guild_id` for "per-route" rate limitation.

    ---

    `channel_id` and `guild_id` are considered as top-level identifiers. On a given
    representation, these will be excluded. You will need to call these values
    separately in order to retrieve their values.
    """

    method: _RouteMethod = field(converter=_RouteMethod)
    """The route's method, e.g. `GET`."""
    path: str = field()
    """The path or URL to the route."""
    channel_id: str = field(default=None)
    """The channel ID associated with this route. This is for route-based rate limits."""
    guild_id: str = field(default=None)
    """The guild ID associated with this route. This is for route-based rate limits."""

    def __str__(self) -> str:
        # We'll be associating our route as just the base URL for the API
        # and the path given. We don't need information for the method unless
        # we're in need of asserting it upon execution via. request.

        # TODO: check to see if the method is truly needed or not.

        return __api_url__ + self.path

    def get_bucket(self, shared: NotNeeded[str] = MISSING) -> str:
        """
        Returns the bucket of the route. This will include top-level identifiers if present.

        Parameters
        ----------
        shared : `str`, optional
            The representation of another bucket as its own route, if present.
            When provided, a shared bucket relationship will be created.

        Returns
        -------
        `str`
            The current bucket of the route.

            If the bucket is not shared, the bucket will be based off of what Discord's
            currently provided to us as the route's information.
        """
        return (
            f"{self.channel_id}:{self.guild_id}:{self.path}"
            if shared is MISSING
            else f"{self.channel_id}:{self.guild_id}:{shared}"
        )


@define(slots=False)
class _Limit:
    """Represents a bucket that exists for a route."""

    event: Event = field(default=Event())
    """The asynchronous event associated to the bucket, used for blocking conditions."""
    reset_after: float = field(default=0.0)
    """The time remaining before the event may be reset. Defaults to `0.0`."""


class HTTPProtocol(Protocol):
    def __init__(self, token: str):
        ...

    async def request(self, route: _Route, payload: dict, retries: NotNeeded[int] = MISSING):
        ...


class HTTPClient(HTTPProtocol):
    """
    Represents a connection to Discord's REST API. The most common use case
    of the Discord API will be providing a service, or access to a platform
    through the OAuth2 API.

    Attributes
    ----------
    token : `str`
        The bot's token.
    _headers : `dict[str]`
        HTTP headers sent for authorisation purposes.
    _limits : `dict[_Route, str]`
        The buckets currently stored.
    _rate_limits : `dict[str, _Limit]`
        The rate limits currently stored.
    _global_rate_limit : `_Limit`
        The global rate limit state.
    """

    __slots__ = ("token", "_headers")
    token: str
    """The bot's token."""
    _headers: dict[str, str]
    """HTTP headers sent for authorisation purposes."""
    _buckets: dict[_Route, str] = {}
    """The buckets currently stored."""
    _rate_limits: dict[str, _Limit] = {}
    """The rate limits currently stored."""
    _global_rate_limit: _Limit = _Limit()
    """The global rate limit state.
"""

    def __init__(self, token: str):
        """
        Creates a new connection to the REST API.

        Parameters
        ----------
        token : `str`
            The bot's token to connect with.
        """
        self.token = token
        self._headers = {
            "Authorization": f"Bot {self.token}",
            "Content-Type": "application/json",
            "User-Agent": f"DiscordBot ({__repo_url__} {__version__}) "
            f"Python/{version_info[0]}.{version_info[1]} "
            f"httpx/{__http_version__}",
        }

    async def request(
        self, route: _Route, payload: dict | None = None, retries: NotNeeded[int] = MISSING
    ):
        """
        Makes a request to Discord's REST API.

        Parameters
        ----------
        route : `_Route`
            The route of the API endpoint.
        payload : `dict`
            The payload to send with the request. If you're making
            a `GET` or `DELETE` call, this dictionary is automatically
            converted into a query-parameter string.
        retries : `int`, optional
            The amount of retries to make if an HTTP request fails.
            Defaults to `1`.

        Returns
        -------
        `dict`
            The JSON associated with the HTTP request.

        Raises: `HTTPException`
        """

        # TODO: Allow a reason field to be passed for audit log
        # purposes.

        if self._global_rate_limit.event.is_set() and self._global_rate_limit.reset_after != 0:
            logger.warning(
                f"There is still a global rate limit ongoing. Trying again in {self._global_rate_limit.reset_after}s."
            )
            await self._global_rate_limit.event.wait(self._global_rate_limit.reset_after)
            self._global_rate_limit.reset_after = 0.0
        elif self._global_rate_limit.reset_after == 0.0:
            self._global_rate_limit.event = Event()

        bucket_path = route.get_bucket()
        rate_limit = self._rate_limits.get(bucket_path)
        if rate_limit:
            if rate_limit.event.is_set() and rate_limit.reset_after != 0:
                logger.warning(
                    f"The current bucket {bucket_path} is still under a rate limit. Trying again in {rate_limit.reset_after}s."
                )
                await rate_limit.event.wait(rate_limit.reset_after)
                rate_limit.reset_after = 0.0
            elif rate_limit.reset_after == 0.0:
                rate_limit.event = Event()
        else:
            self._rate_limits[bucket_path] = _Limit()

        retry_attempts = 1 if retries is MISSING else retries
        for attempt in range(retry_attempts):
            try:
                resp: Response

                async with AsyncClient(headers=self._headers) as client:
                    if route.method in (_RouteMethod.POST, _RouteMethod.PUT):
                        resp = await client.request(route.method.value, str(route), json=payload)
                    elif route.method in (_RouteMethod.GET, _RouteMethod.DELETE):
                        resp = await client.request(
                            route.method.value,
                            str(route),
                            params=QueryParams(**payload) if payload else None,
                        )

                    json = resp.json()
                    logger.debug(f"{route.method} {route}: {resp.status_code}")
                    logger.debug(dumps(loads(json), indent=4, sort_keys=True))

                    if isinstance(json, dict) and json.get("errors"):
                        raise HTTPException(json, severity=INFO)
                    if resp.status_code == 429:
                        reset_after = resp.headers.get("X-RateLimit-Reset-After", 0.0)
                        if bool(resp.headers.get("X-RateLimit-Global")):
                            logger.warning(
                                f"A global rate limit has occured. Locking down future requests for {reset_after}s."
                            )
                            self._global_rate_limit.reset_after = reset_after
                            self._global_rate_limit.set()
                        else:
                            logger.warning(
                                f"A route-based rate limit has occured. Locking down future requests for {reset_after}s."
                            )
                            rate_limit.reset_after = reset_after
                            rate_limit.event.set()
                    if resp.headers.get("X-RateLimit-Remaining", 0) == 0:
                        logger.warning(
                            f"We've reached the maximum number of requests possible. Locking down future requests for {reset_after}s."
                        )
                        await sleep(reset_after)

                    return json
            except OSError as err:
                if attempt <= 1 and err.errno in {54, 10054}:
                    await sleep(5)
            except Exception as err:
                logger.info(err)
